"""
Schema module for Skypydb.
"""

from skypydb.schema.schema import TableDefinition
from skypydb.schema.mixins.schema import (
    defineSchema,
    defineTable,
    SysSchema
)
from skypydb.schema.values import (
    Validator,
    value
)

__all__ = [
    "defineSchema",
    "defineTable",
    "SysSchema",
    "TableDefinition",
    "Validator",
    "value"
]



"""
Module containing the BooleanValidator class, which is used by the validator.
"""

from typing import Any
from skypydb.schema.mixins.values import Validator

class BooleanValidator(Validator):
    """
    Validator for boolean values.
    """

    def validate(
        self,
        value: Any
    ) -> bool:
        """
        Check if value is a boolean.
        """

        return isinstance(value, bool)

    def __repr__(self) -> str:
        return "value.boolean()"



"""
Module containing the VSysDelete class, which is used to delete items in the collection.
"""

from typing import (
    Any,
    Dict,
    List,
    Optional
)
from skypydb.security.validation import InputValidator

class VSysDelete:
    def delete(
        self,
        collection_name: str,
        ids: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, str]] = None
    ) -> int:
        """
        Delete items from a collection.

        Args:
            collection_name: Name of the collection
            ids: Optional list of IDs to delete
            where: Optional metadata filter
            where_document: Optional document filter

        Returns:
            Number of items deleted
        """

        collection_name = InputValidator.validate_table_name(collection_name)
        if not self.collection_exists(collection_name):
            raise ValueError(f"Collection '{collection_name}' not found")

        cursor = self.conn.cursor()
        if ids is not None:
            placeholders = ", ".join(["?" for _ in ids])
            cursor.execute(
                f"DELETE FROM [vec_{collection_name}] WHERE id IN ({placeholders})",
                list(ids)
            )
        else:
            # get all items and filter
            items = self._get_all_items(collection_name)
            ids_to_delete = []

            for item in items:
                if self._matches_filters(item, where, where_document):
                    ids_to_delete.append(item["id"])
            if ids_to_delete:
                placeholders = ", ".join(["?" for _ in ids_to_delete])
                cursor.execute(
                    f"DELETE FROM [vec_{collection_name}] WHERE id IN ({placeholders})",
                    ids_to_delete
                )
        deleted_count = cursor.rowcount
        self.conn.commit()
        return deleted_count


"""
Module containing the EmbeddingsFn class, which is used to generate embeddings for a list of texts.
"""

from typing import (
    List,
    Optional
)

class EmbeddingsFn:
    def __init__(
        self,
        dimension: Optional[int] = None
    ) -> None:
        """
        Initialize the EmbeddingsFn with an optional dimension.

        Args:
            dimension: The dimension of the embeddings. If None, it will be inferred from the first embedding.
        """

        self._dimension = dimension

    def _get_embedding(
        self,
        text: str
    ) -> List[float]:
        """
        Get embedding for a single text.
        """

        raise NotImplementedError(
            f"{self.__class__.__name__} must implement `_get_embedding`."
        )

    def embed(
        self,
        texts: List[str]
    ) -> List[List[float]]:
        """
        Generate embeddings for a list of texts.

        Args:
            texts: List of texts to embed

        Returns:
            List of embedding vectors
        """

        embeddings: List[List[float]] = []

        for text in texts:
            embedding = self._get_embedding(text)
            embeddings.append(embedding)
            # cache the dimension from the first embedding
            if self._dimension is None:
                self._dimension = len(embedding)
        return embeddings

    def dimension(
        self
    ) -> Optional[int]:
        """
        Get the embedding dimension.

        Returns: 
            None if no embedding has been generated yet.
        """

        return self._dimension

    def get_dimension(
        self
    ) -> int:
        """
        Get embedding dimension, generating a test embedding if needed.

        Returns:
            The dimension of embeddings produced by this model.
        """

        if self._dimension is None:
            test_embedding = self._get_embedding("test")
            self._dimension = len(test_embedding)
        return self._dimension



"""
Module containing the SysDelete class, which is used to delete a table.
"""

from skypydb.errors import TableNotFoundError

class SysDelete:
    def delete_table(
        self,
        table_name: str
    ) -> None:
        """
        Delete a table and its configuration.

        Args:
            table_name: Name of the table to delete

        Raises:
            TableNotFoundError: If table doesn't exist

        Example:
            client.delete_table("users")
        """

        if not self.db.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")

        self.db.delete_table(table_name)
        self.db.delete_table_config(table_name)


"""
Module containing the SysQuery class, which is used to query the collection for similar items.
"""

from typing import (
    Any,
    Dict,
    Optional,
    List
)

class SysQuery:
    def query(
        self,
        query_embeddings: Optional[List[List[float]]] = None,
        query_texts: Optional[List[str]] = None,
        n_results: int = 10,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, str]] = None,
        include: Optional[List[str]] = None
    ) -> Dict[str, List[List[Any]]]:
        """
        Query the collection for similar items.

        Args:
            query_embeddings: Optional query embedding vectors
            query_texts: Optional query texts (will be embedded)
            n_results: Number of results to return per query (default: 10)
            where: Optional metadata filter to apply before search
            where_document: Optional document content filter
            include: Optional list of fields to include in results
                    (embeddings, documents, metadatas, distances)

        Returns:
            Dictionary with nested lists of results for each query

        Example:
            # Query with text
            results = collection.query(
                query_texts=["What is machine learning?"],
                n_results=5
            )

            # Query with filters
            results = collection.query(
                query_texts=["AI applications"],
                n_results=10,
                where={"category": "technology"}
            )

            # Access results (first query's results)
            for i, doc_id in enumerate(results["ids"][0]):
                print(f"ID: {doc_id}")
                print(f"Distance: {results['distances'][0][i]}")
                print(f"Document: {results['documents'][0][i]}")
        """

        return self._db.query(
            collection_name=self._name,
            query_embeddings=query_embeddings,
            query_texts=query_texts,
            n_results=n_results,
            where=where,
            where_document=where_document,
            include=include
        )


"""
Module containing the Encryption class, which is used to encrypt and decrypt data.
"""

from typing import (
    Dict,
    Any,
    Optional
)
from skypydb.security.encryption import EncryptionManager

class Encryption:
    def __init__(
        self,
        path: str,
        encryption_key: Optional[str] = None,
        salt: Optional[bytes] = None,
        encrypted_fields: Optional[list] = None
    ):
        """
        Initialize the encryption module.

        Args:
            encryption_key: Optional encryption key for data encryption
            salt: Optional salt for encryption key derivation
            encrypted_fields: Optional list of fields to encrypt
        """

        self.encryption_key = encryption_key
        self.salt = salt
        if encryption_key and encrypted_fields is None:
            raise ValueError(
                "encrypted_fields must be explicitly set when encryption_key is provided; "
                "use [] to disable encryption."
            )
        self.encrypted_fields = encrypted_fields if encrypted_fields is not None else []

        # initialize encryption manager
        self._encryption_manager = EncryptionManager(encryption_key=encryption_key, salt=salt)

    def encrypt_data(
        self,
        data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Encrypt sensitive fields in data dictionary.

        Args:
            data: Dictionary containing data to encrypt

        Returns:
            Dictionary with encrypted fields
        """

        if not self._encryption_manager.enabled:
            return data

        # determine which fields to encrypt
        fields_to_encrypt = [
            key for key in data.keys() 
            if key in self.encrypted_fields
        ]
        return self._encryption_manager.encrypt_dict(data, fields_to_encrypt)

    def decrypt_data(
        self,
        data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Decrypt sensitive fields in data dictionary.

        Args:
            data: Dictionary containing encrypted data

        Returns:
            Dictionary with decrypted fields
        """

        if not self._encryption_manager.enabled:
            return data

        # determine which fields to decrypt
        fields_to_decrypt = [
            key for key in data.keys() 
            if key in self.encrypted_fields
        ]
        return self._encryption_manager.decrypt_dict(data, fields_to_decrypt)



"""
Module containing the SysDelete class, which is used to delete items from a collection.
"""

from typing import (
    Any,
    Dict,
    Optional,
    List
)

class SysDelete:
    def delete(
        self,
        ids: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Delete items from the collection.

        Args:
            ids: Optional list of IDs to delete
            where: Optional metadata filter
            where_document: Optional document content filter

        Example:
            # Delete by ID
            collection.delete(ids=["doc1", "doc2"])

            # Delete by metadata filter
            collection.delete(where={"source": "old"})

            # Delete by document content
            collection.delete(where_document={"$contains": "deprecated"})
        """

        # require at least one selector to avoid accidental deletion of all items.
        if ids is None and where is None and where_document is None:
            raise ValueError(
                "delete() requires at least one of 'ids', 'where', or "
                "'where_document' to be provided."
            )

        self._db.delete(
            collection_name=self._name,
            ids=ids,
            where=where,
            where_document=where_document
        )


"""
Security constants module for Skypydb.
"""

import re

# patterns for validation
TABLE_NAME_PATTERN = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_-]*$')
COLUMN_NAME_PATTERN = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')

# maximum lengths
MAX_TABLE_NAME_LENGTH = 64
MAX_COLUMN_NAME_LENGTH = 64
MAX_STRING_LENGTH = 10000

# SQL injection patterns to detect
SQL_INJECTION_PATTERNS = [
    r';\s*DROP\s+TABLE',
    r';\s*DELETE\s+FROM',
    r';\s*UPDATE\s+',
    r';\s*INSERT\s+INTO',
    r'--',
    r'/\*',
    r'\*/',
    r'xp_',
    r'sp_',
    r'EXEC\s*\(',
    r'EXECUTE\s*\(',
    r'UNION\s+SELECT',
    r'INTO\s+OUTFILE',
    r'LOAD_FILE'
]



"""
Encryption module for securing sensitive data in Skypydb.
"""

from typing import Optional
from skypydb.errors import EncryptionError
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from skypydb.security.mixins.encryption import (
    SysManager,
    SysGenerator,
    SysPassword,
    SysEncrypt,
    SysDecrypt
)

class EncryptionManager(
    SysManager,
    SysGenerator,
    SysPassword,
    SysEncrypt,
    SysDecrypt
):
    """
    Manages encryption and decryption of sensitive data using AES-256-GCM.

    Features:
    - AES-256-GCM encryption (authenticated encryption)
    - PBKDF2HMAC key derivation from passwords
    - Secure random nonce generation
    - Base64 encoding for storage compatibility
    """

    def __init__(
        self,
        encryption_key: Optional[str] = None,
        iterations: int = 100000,
        salt: Optional[bytes] = None
    ):
        """
        Initialize encryption manager.

        Args:
            encryption_key: Master encryption key/password. If None, encryption is disabled.
            iterations: Number of PBKDF2HMAC iterations (default: 100000)
            salt: Required, non-empty salt for PBKDF2HMAC when encryption is enabled

        Raises:
            EncryptionError: If cryptography library is not installed
        """

        if encryption_key is not None and not encryption_key.strip():
            raise EncryptionError("encryption_key must be a non-empty string")

        self.enabled = bool(encryption_key)
        self.iterations = iterations
        self._salt = salt
        self._key: Optional[bytes] = None
        if self.enabled:
            if encryption_key == "":
                raise EncryptionError("Encryption key must not be empty.")
            # derive a 256-bit key from the password
            assert encryption_key is not None  # type narrowing for type checker
            self._key = self._derive_key(encryption_key, salt=self._salt)
            self._aesgcm = AESGCM(self._key)



"""
Module containing the Validator class, which is used to validate type values.
"""

from typing import Any

class Validator:
    """
    Base class for type validators.
    """

    def validate(
        self,
        value: Any
    ) -> bool:
        """
        Validate a value.

        Args:
            value: Value to validate

        Returns:
            True if value is valid
        """

        raise NotImplementedError

    def __repr__(self) -> str:
        raise NotImplementedError



"""
tables module.
"""

from .syscreate import SysCreate
from .sysdelete import SysDelete
from .sysget import SysGet
from .audit import AuditTable

__all__ = [
    "SysCreate",
    "SysDelete",
    "SysGet",
    "AuditTable"
]



"""
Embedding provider factory helpers.
"""

from typing import (
    Any,
    Callable,
    List
)

def _validate_remaining_config(
    provider: str,
    config: dict
) -> None:
    """
    Validate that no unsupported config keys are provided.
    """

    if config:
        unsupported_keys = ", ".join(sorted(config.keys()))
        raise ValueError(
            f"Unsupported embedding config keys for provider '{provider}': {unsupported_keys}"
        )

def get_embedding_function(
    provider: str = "ollama",
    **config: Any
) -> Callable[[List[str]], List[List[float]]]:
    """
    Get an embedding function from supported providers.

    Args:
        provider: Embedding provider (ollama, openai, sentence-transformers)
        **config: Provider-specific configuration
    """

    provider = provider.lower().strip().replace("_", "-")

    # ollama provider
    if provider == "ollama":
        from skypydb.embeddings.ollama import OllamaEmbedding

        model = config.pop("model", "mxbai-embed-large")
        base_url = config.pop("base_url", "http://localhost:11434")
        dimension = config.pop("dimension", None)
        _validate_remaining_config(provider, config)
        return OllamaEmbedding(model=model, base_url=base_url, dimension=dimension)

    # openai provider
    if provider == "openai":
        from skypydb.embeddings.openai import OpenAIEmbedding

        api_key = config.pop("api_key", None)
        model = config.pop("model", "text-embedding-3-small")
        base_url = config.pop("base_url", None)
        organization = config.pop("organization", None)
        project = config.pop("project", None)
        timeout = config.pop("timeout", None)
        dimension = config.pop("dimension", None)
        _validate_remaining_config(provider, config)
        return OpenAIEmbedding(
            api_key=api_key,
            model=model,
            base_url=base_url,
            organization=organization,
            project=project,
            timeout=timeout,
            dimension=dimension
        )

    # sentence-transformers provider
    if provider in {"sentence-transformers", "sentence-transformer"}:
        from skypydb.embeddings.sentence_transformers import SentenceTransformerEmbedding

        model = config.pop("model", "all-MiniLM-L6-v2")
        device = config.pop("device", None)
        normalize_embeddings = config.pop("normalize_embeddings", False)
        dimension = config.pop("dimension", None)
        _validate_remaining_config(provider, config)
        return SentenceTransformerEmbedding(
            model=model,
            device=device,
            normalize_embeddings=normalize_embeddings,
            dimension=dimension
        )
    raise ValueError(
        f"Unsupported embedding provider '{provider}'. "
        "Supported providers: ollama, openai, sentence-transformers."
    )


"""
Vector mixins API client module for Skypydb
"""

from skypydb.api.mixins.vector.syscreate import SysCreate
from skypydb.api.mixins.vector.sysget import SysGet
from skypydb.api.mixins.vector.syslist import SysList
from skypydb.api.mixins.vector.sysdelete import SysDelete
from skypydb.api.mixins.vector.utils import Utils

__all__ = [
    "SysCreate",
    "SysGet",
    "SysList",
    "SysDelete",
    "Utils"
]


"""
Module containing the SysGet class, which is used to get items from a collection.
"""

from typing import (
    Any,
    Dict,
    Optional,
    List
)

class SysGet:
    def get(
        self,
        ids: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, str]] = None,
        include: Optional[List[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None
    ) -> Dict[str, List[Any]]:
        """
        Get items from the collection by ID or filter.

        Args:
            ids: Optional list of IDs to retrieve
            where: Optional metadata filter
            where_document: Optional document content filter
            include: Optional list of fields to include
                    (embeddings, documents, metadatas)
            limit: Optional maximum number of results
            offset: Optional offset for pagination

        Returns:
            Dictionary with lists of ids, embeddings, documents, metadatas

        Example:
            # Get by IDs
            results = collection.get(ids=["doc1", "doc2"])

            # Get by filter
            results = collection.get(
                where={"source": "web"},
                include=["documents", "metadatas"]
            )

            # Access results
            for i, doc_id in enumerate(results["ids"]):
                print(f"ID: {doc_id}")
                print(f"Document: {results['documents'][i]}")
        """

        results = self._db.get(
            collection_name=self._name,
            ids=ids,
            where=where,
            where_document=where_document,
            include=include
        )
        # apply limit and offset
        if offset is not None or limit is not None:
            start = offset or 0
            end = (start + limit) if limit else None

            results["ids"] = results["ids"][start:end]
            if results.get("embeddings") is not None:
                results["embeddings"] = results["embeddings"][start:end]
            if results.get("documents") is not None:
                results["documents"] = results["documents"][start:end]
            if results.get("metadatas") is not None:
                results["metadatas"] = results["metadatas"][start:end]
        return results


"""
Module containing the OptionalValidator class, which is used to validate optional values.
"""

from typing import Any
from skypydb.schema.mixins.values import Validator

class OptionalValidator(Validator):
    """
    Validator for optional values (can be None or the wrapped type).
    """

    def __init__(
        self,
        validator: Validator
    ):
        """
        Initialize optional validator.

        Args:
            validator: The validator for the non-null type
        """

        self.validator = validator
        self.optional = True

    def validate(
        self,
        value: Any
    ) -> bool:
        """
        Check if value is None or valid according to wrapped validator.
        """

        if value is None:
            return True
        return self.validator.validate(value)

    def __repr__(self) -> str:
        return f"value.optional({self.validator})"



"""
Vector database backend using SQLite for storing and querying embeddings.
"""

import sqlite3
from pathlib import Path
from typing import (
    List,
    Optional,
    Callable
)
from skypydb.database.mixins.vector import (
    SysEmbeddings,
    SysAdd,
    SysUpdate,
    SysQuery,
    VSysGet,
    VSysDelete,
    AuditCollections,
    SysCreate,
    SysGet,
    SysCount,
    SysDelete
)

class VectorDatabase(
    SysEmbeddings,
    SysAdd,
    SysUpdate,
    SysQuery,
    VSysGet,
    VSysDelete,
    AuditCollections,
    SysCreate,
    SysGet,
    SysCount,
    SysDelete
):
    """
    Manages SQLite database for vector storage and similarity search.
    """

    def __init__(
        self,
        path: str,
        embedding_function: Optional[Callable[[List[str]], List[List[float]]]] = None
    ):
        """
        Initialize vector database.

        Args:
            path: Path to SQLite database file
            embedding_function: Optional function to generate embeddings from text
        """

        self.path = path
        self.embedding_function = embedding_function

        # create directory if it doesn't exist
        Path(path).parent.mkdir(parents=True, exist_ok=True)

        # connect to SQLite database
        self.conn = sqlite3.connect(path, check_same_thread=False)
        self.conn.row_factory = sqlite3.Row

        # create collections metadata table
        self._ensure_collections_table()

    def close(self) -> None:
        """
        Close database connection.
        """

        if self.conn:
            self.conn.close()



"""
Module containing the SysAdd class, which is used to add items in the collection.
"""

import json
from datetime import datetime
from typing import (
    Dict,
    Any,
    List,
    Optional
)
from skypydb.security.validation import InputValidator

class SysAdd:
    def add(
        self,
        collection_name: str,
        ids: List[str],
        embeddings: Optional[List[List[float]]] = None,
        documents: Optional[List[str]] = None,
        metadatas: Optional[List[Dict[str, Any]]] = None
    ) -> List[str]:
        """
        Add items to a collection.

        Args:
            collection_name: Name of the collection
            ids: List of unique IDs for each item
            embeddings: Optional list of embedding vectors
            documents: Optional list of documents (will be embedded if embedding_function is set)
            metadatas: Optional list of metadata dictionaries
            
        Returns:
            List of IDs of added items
            
        Raises:
            ValueError: If neither embeddings nor documents are provided
        """

        collection_name = InputValidator.validate_table_name(collection_name)
        if not self.collection_exists(collection_name):
            raise ValueError(f"Collection '{collection_name}' not found")
        if embeddings is None and documents is None:
            raise ValueError("Either embeddings or documents must be provided")
        if embeddings is None:
            if self.embedding_function is None:
                raise ValueError(
                    "Documents provided but no embedding function set. "
                    "Either provide embeddings directly or set an embedding_function."
                )
            if documents is None:
                raise ValueError("Either embeddings or documents must be provided")
            embeddings = self.embedding_function(documents)

        # validate lengths match
        n_items = len(ids)
        if len(embeddings) != n_items:
            raise ValueError(
                f"Number of embeddings ({len(embeddings)}) doesn't match "
                f"number of IDs ({n_items})"
            )
        if documents is not None and len(documents) != n_items:
            raise ValueError(
                f"Number of documents ({len(documents)}) doesn't match "
                f"number of IDs ({n_items})"
            )
        if metadatas is not None and len(metadatas) != n_items:
            raise ValueError(
                f"Number of metadatas ({len(metadatas)}) doesn't match "
                f"number of IDs ({n_items})"
            )

        cursor = self.conn.cursor()
        
        now = datetime.now().isoformat()

        for i, item_id in enumerate(ids):
            embedding = embeddings[i]
            document = documents[i] if documents else None
            metadata = metadatas[i] if metadatas else None

            cursor.execute(
                f"""
                INSERT OR REPLACE INTO [vec_{collection_name}] 
                (id, document, embedding, metadata, created_at)
                VALUES (?, ?, ?, ?, ?)
                """,
                (
                    item_id,
                    document,
                    json.dumps(embedding),
                    json.dumps(metadata) if metadata else None,
                    now
                )
            )
        self.conn.commit()
        return ids


"""
Type validators for Skypydb schema.
"""

from skypydb.schema.mixins.values import (
    Validator,
    StringValidator,
    Int64Validator,
    Float64Validator,
    BooleanValidator,
    OptionalValidator
)

class Values:
    """
    Factory for creating type validators.
    """

    # create a string validator
    @staticmethod
    def string() -> Validator:
        """
        Create a string validator.
        """

        return StringValidator()


    # create an integer validator
    @staticmethod
    def int64() -> Validator:
        """
        Create an integer validator.
        """

        return Int64Validator()

    # create a float validator
    @staticmethod
    def float64() -> Validator:
        """
        Create a float validator.
        """

        return Float64Validator()

    # create a boolean validator
    @staticmethod
    def boolean() -> Validator:
        """
        Create a boolean validator.
        """

        return BooleanValidator()

    # create an optional validator
    @staticmethod
    def optional(validator: Validator) -> Validator:
        """
        Create an optional validator.

        Args:
            validator: The validator to make optional

        Returns:
            An optional validator
        """

        return OptionalValidator(validator)

# create singleton instance for easy import
value = Values()



"""
Module containing the SysDecrypt class, which is used to decrypt data.
"""

import base64
from typing import Optional
from skypydb.errors import EncryptionError
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from skypydb.security.mixins.encryption import SysPassword

class SysDecrypt:
    def __init__(
        self,
        encryption_key: Optional[str] = None,
        salt: Optional[bytes] = None
    ):
        if encryption_key is not None and not encryption_key.strip():
            raise EncryptionError("encryption_key must be a non-empty string")

        self.enabled = bool(encryption_key)
        self._salt = salt
        self._key: Optional[bytes] = None
        self._password = SysPassword()
        if self.enabled:
            if encryption_key == "":
                raise EncryptionError("Encryption key must not be empty.")
            # derive a 256-bit key from the password
            assert encryption_key is not None  # type narrowing for type checker
            self._key = self._password._derive_key(encryption_key, salt=self._salt)
            self._aesgcm = AESGCM(self._key)

    def decrypt(
        self,
        encrypted_data: str
    ) -> str:
        """
        Decrypt encrypted data.

        Args:
            encrypted_data: Base64-encoded encrypted data with format: nonce|ciphertext

        Returns:
            Decrypted plaintext

        Raises:
            EncryptionError: If decryption fails
        """

        if not self.enabled:
            return encrypted_data

        try:
            # decode from base64
            encrypted_bytes = base64.b64decode(encrypted_data.encode('utf-8'))

            # extract nonce and ciphertext
            nonce = encrypted_bytes[:12]
            ciphertext = encrypted_bytes[12:]

            # decrypt the data
            plaintext = self._aesgcm.decrypt(
                nonce,
                ciphertext,
                None
            )
            return plaintext.decode('utf-8')
        except Exception as e:
            raise EncryptionError(f"Decryption failed: {str(e)}")

    def decrypt_dict(
        self,
        data: dict,
        fields_to_decrypt: Optional[list] = None
    ) -> dict:
        """
        Decrypt specific fields in a dictionary.

        Args:
            data: Dictionary containing encrypted data
            fields_to_decrypt: List of field names to decrypt. If None, decrypts all values.

        Returns:
            Dictionary with decrypted fields
        """

        if not self.enabled:
            return data

        decrypted_data = {}

        for key, value in data.items():
            if fields_to_decrypt is None or key in fields_to_decrypt:
                # decrypt this field
                if isinstance(value, str) and value:
                    try:
                        decrypted_data[key] = self.decrypt(value)
                    except EncryptionError:
                        # if decryption fails, keep original value but they might be unencrypted data
                        decrypted_data[key] = value
                else:
                    decrypted_data[key] = value
            else:
                # don't decrypt this field
                decrypted_data[key] = value
        return decrypted_data



"""
Module containing the SysValidation class, which is used to validate data in the database.
"""

from skypydb.errors import ValidationError
from typing import (
    Optional,
    Dict,
    Any
)
from skypydb.security.constants import (
    MAX_TABLE_NAME_LENGTH,
    MAX_COLUMN_NAME_LENGTH,
    MAX_STRING_LENGTH,
    TABLE_NAME_PATTERN,
    COLUMN_NAME_PATTERN
)
from skypydb.security.mixins.validation.syscheck import SysCheck
from skypydb.security.mixins.validation.syssanitize import SysSanitize

class SysValidation:
    @classmethod
    def __init__(
        cls,
        sys_check: "SysCheck",
        sys_sanitize: "SysSanitize"
    ):
        cls.sys_check = sys_check
        cls.sys_sanitize = sys_sanitize

    @classmethod
    def validate_table_name(
        cls,
        table_name: str
    ) -> str:
        """
        Validate a table name.

        Args:
            table_name: Name of the table to validate

        Returns:
            Validated table name

        Raises:
            ValidationError: If table name is invalid
        """

        if not table_name:
            raise ValidationError("Table name cannot be empty")
        if not isinstance(table_name, str):
            raise ValidationError("Table name must be a string")
        if len(table_name) > MAX_TABLE_NAME_LENGTH:
            raise ValidationError(
                f"Table name too long (max {MAX_TABLE_NAME_LENGTH} characters)"
            )
        if not TABLE_NAME_PATTERN.match(table_name):
            raise ValidationError(
                "Table name must start with a letter or underscore and contain only "
                "alphanumeric characters, underscores, and hyphens"
            )
        # check for SQL injection patterns
        if cls._contains_sql_injection(table_name):
            raise ValidationError("Table name contains potentially dangerous characters")
        return table_name

    @classmethod
    def validate_column_name(
        cls,
        column_name: str
    ) -> str:
        """
        Validate a column name.

        Args:
            column_name: Name of the column to validate

        Returns:
            Validated column name

        Raises:
            ValidationError: If column name is invalid
        """

        if not column_name:
            raise ValidationError("Column name cannot be empty")
        if not isinstance(column_name, str):
            raise ValidationError("Column name must be a string")
        if len(column_name) > MAX_COLUMN_NAME_LENGTH:
            raise ValidationError(
                f"Column name too long (max {MAX_COLUMN_NAME_LENGTH} characters)"
            )
        if not COLUMN_NAME_PATTERN.match(column_name):
            raise ValidationError(
                "Column name must start with a letter or underscore and contain only "
                "alphanumeric characters and underscores"
            )
        # check for SQL injection patterns
        if cls._contains_sql_injection(column_name):
            raise ValidationError("Column name contains potentially dangerous characters")
        return column_name

    @classmethod
    def validate_string_value(
        cls,
        value: str,
        max_length: Optional[int] = None
    ) -> str:
        """
        Validate a string value.

        Args:
            value: String value to validate
            max_length: Optional maximum length (defaults to MAX_STRING_LENGTH)

        Returns:
            Validated string value

        Raises:
            ValidationError: If string value is invalid
        """

        if not isinstance(value, str):
            raise ValidationError("Value must be a string")

        max_len = max_length or MAX_STRING_LENGTH

        if len(value) > max_len:
            raise ValidationError(
                f"String value too long (max {max_len} characters)"
            )
        return value

    @classmethod
    def validate_data_dict(
        cls,
        data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate a dictionary of data.

        Args:
            data: Dictionary containing data to validate

        Returns:
            Validated data dictionary

        Raises:
            ValidationError: If data is invalid
        """

        if not isinstance(data, dict):
            raise ValidationError("Data must be a dictionary")

        validated_data = {}

        for key, value in data.items():
            # validate column name
            validated_key = cls.validate_column_name(key)

            # validate value based on type
            if isinstance(value, str):
                validated_value = cls.sanitize_string(value)
            elif isinstance(value, (int, float, bool)):
                validated_value = value
            elif value is None:
                validated_value = None
            else:
                # convert to string for other types
                validated_value = cls.sanitize_string(str(value))

            validated_data[validated_key] = validated_value
        return validated_data

    @classmethod
    def validate_filter_dict(
        cls,
        filters: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate filter parameters for search/delete operations.

        Args:
            filters: Dictionary containing filter parameters

        Returns:
            Validated filter dictionary

        Raises:
            ValidationError: If filters are invalid
        """

        if not isinstance(filters, dict):
            raise ValidationError("Filters must be a dictionary")

        validated_filters = {}

        for key, value in filters.items():
            # validate column name
            validated_key = cls.validate_column_name(key)
            # validate value(s)
            if isinstance(value, list):
                validated_value = [
                    cls.sanitize_string(str(v)) if not isinstance(v, (int, float, bool, type(None)))
                    else v
                    for v in value
                ]
            elif isinstance(value, str):
                validated_value = cls.sanitize_string(value)
            elif isinstance(value, (int, float, bool)):
                validated_value = value
            elif value is None:
                validated_value = None
            else:
                validated_value = cls.sanitize_string(str(value))

            validated_filters[validated_key] = validated_value
        return validated_filters

    @classmethod
    def validate_config(
        cls,
        config: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate a table configuration dictionary.

        Args:
            config: Configuration dictionary

        Returns:
            Validated configuration

        Raises:
            ValidationError: If configuration is invalid
        """

        if not isinstance(config, dict):
            raise ValidationError("Configuration must be a dictionary")

        validated_config = {}

        for table_name, table_config in config.items():
            # validate table name
            validated_table_name = cls.validate_table_name(table_name)
            if not isinstance(table_config, dict):
                raise ValidationError(
                    f"Configuration for table '{table_name}' must be a dictionary"
                )
            validated_table_config = {}
            for column_name, column_type in table_config.items():
                # validate column name
                validated_column_name = cls.validate_column_name(column_name)
                # validate column type
                valid_types = [str, int, float, bool, "str", "int", "float", "bool", "auto"]

                if column_type not in valid_types:
                    raise ValidationError(
                        f"Invalid type for column '{column_name}': {column_type}. "
                        f"Valid types are: {valid_types}"
                    )
                validated_table_config[validated_column_name] = column_type
            validated_config[validated_table_name] = validated_table_config
        return validated_config

def validate_table_name(
    table_name: str,
    sys_validation: "SysValidation"
) -> str:
    """
    Convenience function to validate a table name.

    Args:
        table_name: Name to validate
        sys_validation: SysValidation instance to use for validation

    Returns:
        Validated table name
    """

    return sys_validation.validate_table_name(table_name)

def validate_column_name(
    column_name: str,
    sys_validation: "SysValidation"
) -> str:
    """
    Convenience function to validate a column name.

    Args:
        column_name: Name to validate
        sys_validation: SysValidation instance to use for validation

    Returns:
        Validated column name
    """

    return sys_validation.validate_column_name(column_name)



"""
Module containing the SysAdd class, which is used to get information about the collection.
"""

import json
from typing import (
    List,
    Optional,
    Dict,
    Any
)
from skypydb.security.validation import InputValidator

class SysGet:
    def get_collection(
        self,
        name: str
    ) -> Optional[Dict[str, Any]]:
        """
        Get collection metadata.

        Args:
            name: Collection name

        Returns:
            Collection metadata or None if not found
        """

        name = InputValidator.validate_table_name(name)
        if not self.collection_exists(name):
            return None

        cursor = self.conn.cursor()
        
        cursor.execute(
            "SELECT * FROM _vector_collections WHERE name = ?",
            (name,)
        )
        row = cursor.fetchone()
        if row:
            return {
                "name": row["name"],
                "metadata": json.loads(row["metadata"]) if row["metadata"] else {},
                "created_at": row["created_at"]
            }
        return None

    def get_or_create_collection(
        self,
        name: str,
        metadata: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Get an existing collection or create a new one.

        Args:
            name: Collection name
            metadata: Optional collection metadata (used only if creating)

        Returns:
            Collection metadata
        """

        name = InputValidator.validate_table_name(name)
        if not self.collection_exists(name):
            self.create_collection(name, metadata)

        result = self.get_collection(name)
        # at this point collection must exist since we just created it if needed
        assert result is not None
        return result

    def list_collections(self) -> List[Dict[str, Any]]:
        """
        List all collections.
        
        Returns:
            List of collection metadata dictionaries
        """

        cursor = self.conn.cursor()
        
        cursor.execute("SELECT * FROM _vector_collections")

        collections = []
        for row in cursor.fetchall():
            collections.append({
                "name": row["name"],
                "metadata": json.loads(row["metadata"]) if row["metadata"] else {},
                "created_at": row["created_at"]
            })
        return collections


"""
Module containing the RSysSearch class, which is used to search tables in the database.
"""

import sqlite3
from typing import (
    Any,
    Dict,
    List,
    Optional
)
from skypydb.errors import TableNotFoundError
from skypydb.security.validation import InputValidator
from skypydb.errors import ValidationError
from skypydb.database.mixins.reactive.tables.audit import AuditTable
from skypydb.database.mixins.reactive.tables.sysget import SysGet
from skypydb.database.mixins.reactive.encryption import Encryption

class RSysSearch:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None,
        encryption: Optional[Encryption] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

        self.audit = AuditTable(conn=self.conn)
        self.sysget = SysGet(conn=self.conn, encryption=encryption)
        self.encryption = encryption

    def search(
        self,
        table_name: str,
        index: Optional[str] = None,
        **filters
    ) -> List[Dict[str, Any]]:
        """
        Search a table for rows that match an optional free text index value and/or column filters.

        Parameters:
            table_name (str): Name of the target table.
            index (Optional[str]): Value to search across all non-standard columns; ignored if None.
            **filters: Column-value pairs to filter results. If a value is a list, it is used with an IN clause (empty lists are invalid).

        Returns:
            List[Dict[str, Any]]: Matching rows as dictionaries with sensitive fields decrypted.

        Raises:
            ValidationError: If table name, index, or filters fail validation, or an empty list is provided for a filter.
            TableNotFoundError: If the specified table does not exist.
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        # validate filters
        if filters:
            filters = InputValidator.validate_filter_dict(filters)
        # sanitize index value
        if index is not None:
            index = InputValidator.sanitize_string(str(index))
        if not self.audit.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")

        conditions = []
        params = []

        # add index condition if provided
        # index searches across all non-standard columns (OR condition)
        if index is not None:
            columns = self.sysget.get_table_columns_names(table_name)
            non_standard_columns = [
                col for col in columns if col not in ("id", "created_at")
            ]
            if non_standard_columns:
                # search index value in any of the non-standard columns
                index_conditions = []
                for col in non_standard_columns:
                    index_conditions.append(f"[{col}] = ?")
                    params.append(str(index))
                conditions.append(f"({' OR '.join(index_conditions)})")
        # add additional filters (AND conditions)
        for column, value in filters.items():
            # handle list values, use IN clause
            if isinstance(value, list):
                if not value:
                    raise ValidationError(f"Empty list provided for filter '{column}'")
                placeholders = ", ".join(["?" for _ in value])
                conditions.append(f"[{column}] IN ({placeholders})")
                params.extend([str(v) for v in value])
            else:
                conditions.append(f"[{column}] = ?")
                params.append(str(value))

        # build query
        where_clause = " AND ".join(conditions) if conditions else "1=1"
        query = f"SELECT * FROM [{table_name}] WHERE {where_clause}"

        cursor = self.conn.cursor()

        cursor.execute(query, params)

        # convert rows to dictionaries and decrypt sensitive data
        results = []
        for row in cursor.fetchall():
            row_dict = dict(row)
            if self.encryption:
                decrypted_row = self.encryption.decrypt_data(row_dict)
                results.append(decrypted_row)
            else:
                results.append(row_dict)
        return results



"""
Cli for Skypydb.
"""

from io import BytesIO
from pathlib import (
    Path,
    PurePosixPath
)
from urllib.error import (
    HTTPError,
    URLError
)
from urllib.request import (
    Request,
    urlopen
)
from zipfile import ZipFile
import subprocess
import shutil
import typer
import questionary
from rich import print
from skypydb.security import EncryptionManager
from skypydb import __version__

app = typer.Typer(
    name="skypydb",
    help="Skypydb CLI - Open Source Reactive and Vector Embedding Database"
)

class SkypyCLI:
    """
    Skypy CLI class to manage CLI operations.
    """

    def __init__(
        self,
        env_file_name: str = ".env.local",
        skypydb_folder: str = "db",
        generated_folder: str = "_generated",
        schema_file_name: str = "schema.py",
        repo_zip_url: str = "https://github.com/Ahen-Studio/the-skypydb-dashboard/archive/refs/heads/main.zip",
        repo_dashboard_path: str = "dashboard",
        gitignore_path: str = ".gitignore",
        gitignore_entry: str = ".env.local",
        cwd: Path = Path.cwd()
    ):
        """
        Initialize the CLI with configuration variables.
        """

        self.env_file_name = env_file_name
        self.skypydb_folder = skypydb_folder
        self.generated_folder = generated_folder
        self.schema_file_name = schema_file_name
        self.repo_zip_url = repo_zip_url
        self.repo_dashboard_path = repo_dashboard_path.strip("/\\")
        self.gitignore_path = gitignore_path
        self.gitignore_entry = gitignore_entry
        self.cwd = cwd

    def _find_npm(self):
        """
        Locate the npm executable, handling common platform-specific names.
        """

        return shutil.which("npm") or shutil.which("npm.cmd")

    def launch_dashboard(
        self,
        api_port: int = 8000,
        dashboard_port: int = 3000
    ) -> None:
        """
        Launch the FastAPI server and the dashboard dev server.
        """

        dashboard_dir = self.cwd / self.skypydb_folder / self.generated_folder / self.repo_dashboard_path
        if not dashboard_dir.exists():
            print(
                f"[yellow]Dashboard folder not found at {dashboard_dir}. "
                "Run project initialization first.[/yellow]"
            )
            return

        api_command = [
            "python",
            "-c",
            (
                "import uvicorn; "
                "import skypydb.server.fastapi.server as server; "
                f"uvicorn.run(server.app, host='0.0.0.0', port={api_port})"
            ),
        ]

        npm_path = self._find_npm()
        if not npm_path:
            print("[yellow]npm was not found in PATH. Please install Node.js/npm and try again.[/yellow]")
            return

        dashboard_command = [
            npm_path,
            "run",
            "dev",
            "--",
            "--port",
            str(dashboard_port)
        ]

        print(f"[green]Starting API server on port {api_port}[/green]")
        api_process = subprocess.Popen(
            api_command,
            cwd=str(self.cwd)
        )
        dashboard_process = None

        try:
            next_dir = dashboard_dir / "node_modules" / ".bin"
            next_bin = next_dir / "next"
            next_cmd = next_dir / "next.cmd"
            if not next_bin.exists() and not next_cmd.exists():
                print("[yellow]Installing dashboard dependencies.[/yellow]")
                install_process = subprocess.Popen(
                    [
                        npm_path,
                        "install"
                    ],
                    cwd=str(dashboard_dir)
                )
                install_process.wait()
                if install_process.returncode != 0:
                    print(
                        "[red]Failed to install dashboard dependencies. "
                        "Please check the npm output above and try again.[/red]"
                    )
                    if api_process.poll() is None:
                        api_process.terminate()
                    return

            print(f"[green]Starting dashboard on port {dashboard_port}[/green]")
            dashboard_process = subprocess.Popen(
                dashboard_command,
                cwd=str(dashboard_dir)
            )
        except FileNotFoundError as exc:
            print(f"[yellow]Failed to start dashboard: {exc}[/yellow]")
            api_process.terminate()
            return

        try:
            if dashboard_process:
                api_process.wait()
                dashboard_process.wait()
            else:
                api_process.wait()
        except KeyboardInterrupt:
            print("\n[yellow]Shutting down servers.[/yellow]")
        finally:
            if api_process.poll() is None:
                api_process.terminate()
            if dashboard_process and dashboard_process.poll() is None:
                dashboard_process.terminate()

    def init_project(self) -> None:
        """
        Initialize project with encryption keys and project structure.
        """

        # create project structure
        self._create_project_structure()

        # generate, save encryption keys and update .gitignore
        self._generate_encryption_keys_and_update_gitignore()

        # download the dashboard folder from GitHub
        self._download_dashboard_folder()

        print(f"Write your Skypydb functions in {Path(self.skypydb_folder) / self.schema_file_name}")
        print("Give us feedback at https://github.com/Ahen-Studio/skypy-db/issues")

    def _create_project_structure(self) -> None:
        """
        Create the project directory structure.
        """

        # create db folder
        skypydb_dir = self.cwd / self.skypydb_folder
        if not skypydb_dir.exists():
            skypydb_dir.mkdir(exist_ok=True)

        # create _generated folder
        generated_dir = skypydb_dir / self.generated_folder
        if not generated_dir.exists():
            generated_dir.mkdir(exist_ok=True)

        # create schema.py file
        schema_file = skypydb_dir / self.schema_file_name
        if not schema_file.exists():
            schema_file.write_text("", encoding="utf-8")

        print("[green]✔ Initialized project[/green]")

    def _generate_encryption_keys_and_update_gitignore(self) -> None:
        """
        Generate and save encryption keys to .env.local and update the gitignore file to untrack the .env.local file.
        """

        # generate encryption key and encode the salt key with base64
        encryption_key = EncryptionManager.generate_key()
        salt_key = EncryptionManager.generate_salt()

        # save encryption key and salt key to .env.local
        env_path = self.cwd / self.env_file_name
        content = f"ENCRYPTION_KEY={encryption_key}\nSALT_KEY={salt_key}\n"
        env_path.write_text(content, encoding="utf-8")

        # generate .gitignore if not exists
        gitignore_path = self.cwd / self.gitignore_path
        content = gitignore_path.read_text(encoding="utf-8") if gitignore_path.exists() else ""

        # add .env.local to .gitignore if not exists
        if self.gitignore_entry not in content.splitlines():
            content = f"{content.rstrip()}\n{self.gitignore_entry}\n" if content else f"{self.gitignore_entry}\n"
            gitignore_path.write_text(content, encoding="utf-8")

        print("[green]✔ Encryption keys provided and saved:[/green]")
        print("    encryption key as ENCRYPTION_KEY")
        print("    salt key as SALT_KEY")
        print(f"to {self.env_file_name}")
        print(f'[bold bright_black]Added "{self.env_file_name}" to {self.gitignore_path}[/bold bright_black]\n')

    def _download_dashboard_folder(self) -> None:
        """
        Download the dashboard folder from the GitHub repo zip into the generated directory.
        """

        if not self.repo_zip_url or not self.repo_dashboard_path:
            return

        generated_dir = self.cwd / self.skypydb_folder / self.generated_folder
        target_root = generated_dir / self.repo_dashboard_path

        try:
            request = Request(
                self.repo_zip_url,
                headers={"User-Agent": "skypydb-cli"}
            )
            with urlopen(request) as response:
                zip_bytes = BytesIO(response.read())
        except (HTTPError, URLError) as exc:
            print(f"[yellow]Warning: failed to download dashboard folder. {exc}[/yellow]")
            return

        created_files = 0
        skipped_files = 0

        with ZipFile(zip_bytes) as archive:
            names = archive.namelist()
            repo_root = names[0].split("/")[0] if names else None
            if not repo_root:
                print("[yellow]Warning: unexpected zip format (missing root).[/yellow]")
                return
            prefix = f"{repo_root}/{self.repo_dashboard_path}/"
            for info in archive.infolist():
                if not info.filename.startswith(prefix):
                    continue
                relative = info.filename[len(prefix):]
                if not relative:
                    continue
                rel_path = PurePosixPath(relative)
                if ".." in rel_path.parts:
                    continue
                target_path = target_root / Path(*rel_path.parts)
                if info.is_dir():
                    target_path.mkdir(parents=True, exist_ok=True)
                elif target_path.exists():
                    skipped_files += 1
                else:
                    target_path.parent.mkdir(parents=True, exist_ok=True)
                    with archive.open(info) as src:
                        target_path.write_bytes(src.read())
                    created_files += 1
        if created_files:
            print(f"[green]✔ Downloaded dashboard folder to {target_root}[/green]")
        if skipped_files:
            print(f"[yellow]Skipped {skipped_files} existing file(s) in {target_root}[/yellow]")

@app.command()
def dev() -> None:
    """
    Start Skypydb interactive mode.
    """

    cli = SkypyCLI()

    choices = [
        questionary.Choice(title="create a new project", value="create"),
        questionary.Choice(title="launch the dashboard", value="dashboard"),
        questionary.Choice(title="no thanks", value="exit"),
    ]

    selection = questionary.select(
        "Welcome to Skypydb! Would you like to create a new project?",
        choices=choices,
        qmark="?",
        pointer="❯",
    ).ask()

    if selection is None or selection == "exit":
        print("\n[yellow]Exiting.[/yellow]")
        raise typer.Exit(code=0)
    if selection == "create":
        cli.init_project()
    elif selection == "dashboard":
        cli.launch_dashboard()

def _version_callback(value: bool) -> None:
    """
    Show the skypydb version and exit.
    """

    if value:
        print(f"skypydb {__version__}")
        raise typer.Exit()

@app.callback()
def main_callback(
    version: bool = typer.Option(
        False,
        "--version",
        help="Show version and exit",
        is_eager=True,
        callback=_version_callback
    )
) -> None:
    """
    Main cli callback.
    """

    pass

def main() -> None:
    """
    Main entry point for the CLI.
    """

    app()

if __name__ == "__main__":
    main()



"""
Module containing the SysEmbeddings class, which is used to set the embedding function for the vector datebase.
"""

from typing import (
    List,
    Callable
)

class SysEmbeddings:
    def set_embedding_function(
        self,
        embedding_function: Callable[[List[str]], List[List[float]]]
    ) -> None:
        """
        Set the embedding function for the database.

        Args:
            embedding_function: Function that takes texts and returns embeddings
        """

        self.embedding_function = embedding_function


"""
Module containing the SysDelete class, which is used to delete tables in the database.
"""

import sqlite3
from typing import Optional
from skypydb.security.validation import InputValidator
from skypydb.errors import TableNotFoundError
from skypydb.database.mixins.reactive.tables.audit import AuditTable
from skypydb.database.mixins.reactive.utils import Utils

class SysDelete:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

        self.audit = AuditTable(conn=self.conn)
        self.utils = Utils(conn=self.conn)

    def delete_table(
        self,
        table_name: str
    ) -> None:
        """
        Delete a table.

        Args:
            table_name: Name of the table to delete

        Raises:
            TableNotFoundError: If table doesn't exist
            ValidationError: If table name is invalid
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)
        if not self.audit.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")

        cursor = self.conn.cursor()

        cursor.execute(f"DROP TABLE [{table_name}]")

        self.utils.delete_table_config(table_name)

        self.conn.commit()



"""
Module containing the SysCreate class, which is used to create tables in the database.
"""

import sqlite3
from typing import Optional
from skypydb.errors import TableAlreadyExistsError
from skypydb.security.validation import InputValidator
from skypydb.schema.schema import TableDefinition
from skypydb.database.mixins.reactive.tables.audit import AuditTable
from skypydb.database.mixins.reactive.utils import Utils

class SysCreate:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

        self.audit = AuditTable(conn=self.conn)
        self.utils = Utils(conn=self.conn)

    def create_table(
        self,
        table_name: str,
        table_def: TableDefinition
    ) -> None:
        """
        Create a table based on a TableDefinition from the schema system.

        Args:
            table_name: Name of the table to create
            table_def: TableDefinition containing columns and indexes

        Raises:
            TableAlreadyExistsError: If table already exists
            ValidationError: If table definition is invalid

        Example:
            table_def = defineTable({
                "name": value.string(),
                "email": value.string()
            })
            .index("by_email", ["email"])

            database.create_table("users", table_def)
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        # validate column names
        for col_name in table_def.columns.keys():
            InputValidator.validate_column_name(col_name)
        if self.audit.table_exists(table_name):
            raise TableAlreadyExistsError(f"Table '{table_name}' already exists")

        cursor = self.conn.cursor()

        # get SQL column definitions from table definition
        sql_columns = table_def.get_sql_columns()
        columns_sql = ", ".join(sql_columns)

        # create table
        cursor.execute(
            f"""
            CREATE TABLE [{table_name}] (
                {columns_sql}
            )
            """
        )
        # create indexes
        for index_sql in table_def.get_sql_indexes():
            cursor.execute(index_sql)

        # save table definition as configuration
        config = self.utils.table_def_to_config(table_def)
        self.utils.save_table_config(table_name, config)
        self.conn.commit()



"""
API module.
"""

from .reactive_client import ReactiveClient
from .vector_client import VectorClient
from .collection import Collection

__all__ = [
    "ReactiveClient",
    "VectorClient",
    "Collection"
]



"""
Module containing the SysGet class, which is used to get all the data from a table in the database.
"""

from typing import (
    List,
    Dict,
    Any
)
from skypydb.database.reactive_db import ReactiveDatabase

class SysGet:
    def __init__(
        self,
        db: "ReactiveDatabase",
        table_name: str
    ):
        self.db = db
        self.table_name = table_name

    def get_all(
        self
    ) -> List[Dict[str, Any]]:
        """
        Get all data from the table.
        """

        return self.db.get_all_data(self.table_name)



"""
Module containing the SysSchema class, which is used to perform get operations on the schema in the database.
"""

from typing import (
    List,
    Dict,
    Optional,
    TYPE_CHECKING
)

if TYPE_CHECKING:
    from skypydb.schema.schema import TableDefinition

class SysSchema:
    def __init__(
        self,
        tables: Dict[str, "TableDefinition"]
    ):
        self.tables = tables

    def get_table_definition(
        self,
        table_name: str
    ) -> Optional["TableDefinition"]:
        """
        Get a table definition by name.

        Args:
            table_name: Name of the table

        Returns:
            TableDefinition if found, None otherwise
        """

        return self.tables.get(table_name)

    def get_all_table_names(self) -> List[str]:
        """
        Get all table names in the schema.

        Returns:
            List of table names
        """

        return list(self.tables.keys())



"""
Security module for Skypydb.
"""

from skypydb.security.encryption import EncryptionManager, EncryptionError
from skypydb.security.mixins.encryption import create_encryption_manager
from skypydb.security.mixins.validation import (
    validate_table_name,
    validate_column_name,
    sanitize_input
)
from skypydb.security.validation import InputValidator

__all__ = [
    "create_encryption_manager",
    "EncryptionError",
    "EncryptionManager",
    "InputValidator",
    "sanitize_input",
    "validate_column_name",
    "validate_table_name"
]



"""
Module containing the AuditTable class, which is used to check table operations in the database.
"""

import sqlite3
from typing import (
    Any,
    Dict,
    List,
    Optional
)
from skypydb.security.validation import InputValidator
from skypydb.errors import TableNotFoundError, ValidationError
from skypydb.database.mixins.reactive.utils import Utils

class AuditTable:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

        self.utils = Utils(conn=self.conn)

    def table_exists(
        self,
        table_name: str
    ) -> bool:
        """
        Check if a table exists.
        """

        # validate table name
        try:
            table_name = InputValidator.validate_table_name(table_name)
        except ValidationError:
            return False

        cursor = self.conn.cursor()

        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name=?",
            (table_name,),
        )
        return cursor.fetchone() is not None

    def add_columns_if_needed(
        self,
        table_name: str,
        columns: List[str]
    ) -> None:
        """
        Add columns to a table if they don't exist.

        Args:
            table_name: Name of the table
            columns: List of column names to add
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        existing_columns = set(self.get_table_columns(table_name))

        cursor = self.conn.cursor()

        for column in columns:
            # validate column name
            validated_column = InputValidator.validate_column_name(column)
            if validated_column not in existing_columns and validated_column not in ("id", "created_at"):
                cursor.execute(f"ALTER TABLE [{table_name}] ADD COLUMN [{validated_column}] TEXT")

        self.conn.commit()

    def get_table_columns(
        self,
        table_name: str
    ) -> List[str]:
        """
        Get list of column names for a table.
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        if not self.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")

        cursor = self.conn.cursor()

        cursor.execute(f"PRAGMA table_info([{table_name}])")
        return [row[1] for row in cursor.fetchall()]

    def check_config_table(self) -> None:
        """
        Create the system table for storing table configurations if it doesn't exist.
        """

        cursor = self.conn.cursor()

        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS _skypy_config (
                table_name TEXT PRIMARY KEY,
                config TEXT NOT NULL,
                created_at TEXT NOT NULL
            )
            """
        )
        self.conn.commit()

    def validate_data_with_config(
        self,
        table_name: str,
        data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate data against the table's configuration, converts values to the correct type based on configuration.

        Args:
            table_name: Name of the table
            data: Data dictionary to validate

        Returns:
            Validated data dictionary with converted values

        Raises:
            ValueError: If data validation fails
        """

        config = self.utils.get_table_config(table_name)
        if not config:
            # no configuration, return data as-is
            return data

        validated_data = {}

        for key, value in data.items():
            if key in config:
                expected_type = config[key]
                optional = False
                if isinstance(expected_type, dict):
                    optional = expected_type.get("optional", False)
                    expected_type = expected_type.get("type", "str")

                if value is None and optional:
                    validated_data[key] = None
                    continue
                # skip "auto" type
                if expected_type == "auto" or expected_type == "id":
                    continue
                # type conversion and validation
                if expected_type is str or expected_type == "str":
                    validated_data[key] = str(value)
                elif expected_type is int or expected_type == "int":
                    try:
                        validated_data[key] = int(value)
                    except (ValueError, TypeError):
                        raise ValueError(
                            f"Invalid type for column '{key}': expected int"
                        )
                elif expected_type is float or expected_type == "float":
                    try:
                        validated_data[key] = float(value)
                    except (ValueError, TypeError):
                        raise ValueError(
                            f"Invalid type for column '{key}': expected float"
                        )
                elif expected_type is bool or expected_type == "bool":
                    if isinstance(value, str):
                        validated_data[key] = value.lower() in ("true", "1", "yes")
                    else:
                        validated_data[key] = bool(value)
                else:
                    # unknown type, store as string
                    validated_data[key] = str(value)
            else:
                # column not in config, store as-is
                validated_data[key] = value
        return validated_data



"""
Module containing the SysSearch class, which is used to search data from a table in the database.
"""

from typing import (
    Optional,
    List,
    Dict,
    Any
)
from skypydb.database.reactive_db import ReactiveDatabase

class SysSearch:
    def __init__(
        self,
        db: "ReactiveDatabase",
        table_name: str
    ):
        self.db = db
        self.table_name = table_name

    def search(
        self,
        index: Optional[str] = None,
        **filters
    ) -> List[Dict[str, Any]]:
        """
        Search for data in the table.

        Args:
            index: Value to search for in the index column (primary search key)
            **filters: Additional filters as keyword arguments (column name = value or list of values)

        Returns:
            List of dictionaries containing matching rows

        Example:
            # Search by index and a single filter
            results = table.search(
                index="user123",
                title="document"
            )

            # Search with multiple criteria
            results = table.search(
                index="user123",
                status="active",
                category="news",
            )

            # Search with list values (e.g. uses IN clause in underlying DB)
            results = table.search(
                index="user123",
                title=["doc1", "doc2"]
            )
        """

        # pass filters directly; list values are handled explicitly by the database layer
        return self.db.search(self.table_name, index=index, **filters)



"""
Module containing the SysCount class, which is used to count items in a collection.
"""

from skypydb.security.validation import InputValidator

class SysCount:
    def count(
        self,
        collection_name: str
    ) -> int:
        """
        Count items in a collection.

        Args:
            collection_name: Name of the collection

        Returns:
            Number of items in the collection
        """

        collection_name = InputValidator.validate_table_name(collection_name)

        if not self.collection_exists(collection_name):
            raise ValueError(f"Collection '{collection_name}' not found")

        cursor = self.conn.cursor()

        cursor.execute(f"SELECT COUNT(*) FROM [vec_{collection_name}]")
        return cursor.fetchone()[0]


"""
Dashboard API for monitoring Skypydb databases.
"""

import os
from pathlib import Path
import time
from typing import (
    Dict,
    List,
    Optional,
    Any
)
from dataclasses import dataclass
from skypydb.database.reactive_db import ReactiveDatabase
from skypydb.database.vector_db import VectorDatabase
from skypydb.database.database_linker import DatabaseLinker

@dataclass
class TableInfo:
    """
    Information about a database table.
    """

    name: str
    row_count: int
    columns: List[str]
    config: Optional[Dict] = None

@dataclass
class VectorCollectionInfo:
    """
    Information about a vector collection.
    """

    name: str
    document_count: int
    metadata: Dict[str, Any]

@dataclass
class PaginatedResult:
    """
    Paginated query result.
    """

    data: List[Dict]
    total: int
    limit: int
    offset: int
    has_more: bool

class DatabaseConnection:
    """
    Manages database connections.
    """

    @staticmethod
    def _resolve_db_path(
        env_key: str,
        default_relative: str,
        db_type: str
    ) -> str:
        """
        Resolve a database path from env, normalizing to an absolute path.
        """

        base = Path.cwd()
        raw = os.environ.get(env_key)
        if raw:
            path = Path(raw)
            return str(path if path.is_absolute() else (base / path).resolve())

        discovered = DatabaseConnection.discover_links()
        for item in discovered:
            if item.get("type") != db_type:
                continue
            candidate = Path(item.get("path", ""))
            if candidate.exists():
                return str(candidate.resolve())

        default_path = (base / default_relative).resolve()
        if default_path.exists():
            return str(default_path)

        generated_dir = (base / "db" / "_generated")
        if generated_dir.exists():
            candidates = sorted(p for p in generated_dir.glob("*.db") if p.is_file())
            if candidates:
                return str(candidates[0].resolve())
        return str(default_path)

    @staticmethod
    def discover_links() -> List[Dict[str, str]]:
        """
        Discover database link metadata from current working directory.
        """

        database_linker = DatabaseLinker()
        return database_linker.discover_database_links(Path.cwd())

    @staticmethod
    def _require_existing(path: str, label: str) -> None:
        """
        Ensure the database file exists before connecting.
        """

        if not Path(path).exists():
            raise FileNotFoundError(
                f"{label} database file not found at: {path}. "
                "Set the correct path with the request headers or env vars."
            )

    @staticmethod
    def get_main() -> ReactiveDatabase:
        """
        Get main database instance from environment.
        """

        path = DatabaseConnection._resolve_db_path(
            "SKYPYDB_PATH",
            "db/_generated/skypydb.db",
            "reactive"
        )
        DatabaseConnection._require_existing(path, "Main")
        return ReactiveDatabase(path)

    @staticmethod
    def get_vector() -> VectorDatabase:
        """
        Get vector database instance from environment.
        """

        path = DatabaseConnection._resolve_db_path(
            "SKYPYDB_VECTOR_PATH",
            "db/_generated/vector.db",
            "vector"
        )
        DatabaseConnection._require_existing(path, "Vector")
        return VectorDatabase(path)

class LinksAPI:
    """
    API for database type/path discovery metadata.
    """

    def list_all(self) -> List[Dict[str, str]]:
        """
        List discovered DB links from the project root.
        """

        return DatabaseConnection.discover_links()

class HealthAPI:
    """
    API for checking system health status.
    """

    def check(self) -> Dict[str, Any]:
        """
        Check health status of all database components.
        
        Returns:
            Dictionary with timestamp, overall status, and database statuses
        """

        status = {
            "timestamp": time.time_ns(),
            "status": "healthy",
            "databases": {}
        }

        self._check_main(status)
        self._check_vector(status)
        return status

    def _check_main(
        self,
        status: Dict[str, Any]
    ) -> None:
        """
        Check main database health.
        """

        try:
            db = DatabaseConnection.get_main()
            table_count = len(db.get_all_tables_names())
            db.close()

            status["databases"]["main"] = {
                "status": "connected",
                "tables": table_count
            }
        except Exception as error:
            status["databases"]["main"] = {
                "status": "error",
                "error": str(error)
            }
            status["status"] = "degraded"

    def _check_vector(
        self,
        status: Dict[str, Any]
    ) -> None:
        """
        Check vector database health.
        """

        vdb = None
        try:
            vdb = DatabaseConnection.get_vector()
            collection_count = len(vdb.list_collections())

            status["databases"]["vector"] = {
                "status": "connected",
                "collections": collection_count
            }
        except Exception as error:
            status["databases"]["vector"] = {
                "status": "error",
                "error": str(error)
            }
            status["status"] = "degraded"
        finally:
            if vdb is not None:
                try:
                    vdb.close()
                except Exception:
                    # suppress close errors in health check to avoid masking original issues
                    pass

class TableAPI:
    """
    API for table operations.
    """

    def list_all(self) -> List[Dict[str, Any]]:
        """
        Get all tables with metadata and row counts.
        """

        db = DatabaseConnection.get_main()
        
        try:
            table_names = db.get_all_tables_names()
            return [self._get_info(db, name) for name in table_names]
        finally:
            db.close()

    def _get_info(
        self,
        db: ReactiveDatabase,
        table_name: str
    ) -> Dict[str, Any]:
        """
        Get information about a specific table.
        """

        try:
            return {
                "name": table_name,
                "row_count": len(db.get_all_data(table_name)),
                "columns": db.get_table_columns(table_name),
                "config": db.get_table_config(table_name)
            }
        except Exception:
            return {
                "name": table_name,
                "row_count": 0,
                "columns": [],
                "config": None
            }

    def get_schema(
        self,
        table_name: str
    ) -> Dict[str, Any]:
        """
        Get schema information for a table.
        """

        db = DatabaseConnection.get_main()

        try:
            return {
                "name": table_name,
                "columns": db.get_table_columns(table_name),
                "config": db.get_table_config(table_name)
            }
        finally:
            db.close()

    def get_data(
        self,
        table_name: str,
        limit: int = 100,
        offset: int = 0
    ) -> Dict[str, Any]:
        """
        Get paginated data from a table.
        """

        db = DatabaseConnection.get_main()

        try:
            all_data = db.get_all_data(table_name)
            return self._paginate(all_data, limit, offset)
        finally:
            db.close()

    def search(
        self,
        table_name: str,
        query: Optional[str] = None,
        limit: int = 100,
        **filters
    ) -> Dict[str, Any]:
        """
        Search table data with filters.
        """

        db = DatabaseConnection.get_main()

        try:
            results = db.search(table_name, index=query, **filters)
            if limit and len(results) > limit:
                results = results[:limit]
            return {
                "data": results,
                "total": len(results),
                "limit": limit
            }
        finally:
            db.close()

    def _paginate(
        self,
        data: List[Dict],
        limit: int,
        offset: int
    ) -> Dict[str, Any]:
        """
        Apply pagination to data.
        """

        total = len(data)
        start = offset
        end = offset + limit if limit else total
        return {
            "data": data[start:end],
            "total": total,
            "limit": limit,
            "offset": offset,
            "has_more": end < total
        }

class VectorAPI:
    """
    API for vector collection operations.
    """

    def list_all(self) -> List[Dict[str, Any]]:
        """
        Get all vector collections with document counts.
        """

        vdb = DatabaseConnection.get_vector()

        try:
            collections = vdb.list_collections()
            return [self._get_info(vdb, coll) for coll in collections]
        except Exception:
            return []
        finally:
            vdb.close()

    def _get_info(
        self,
        vdb: VectorDatabase,
        collection: Dict
    ) -> Dict[str, Any]:
        """
        Get information about a vector collection.
        """

        name = collection['name']
        
        try:
            return {
                "name": name,
                "document_count": vdb.count(name),
                "metadata": collection.get('metadata', {})
            }
        except Exception:
            return {
                "name": name,
                "document_count": 0,
                "metadata": collection.get('metadata', {})
            }

    def get_details(
        self,
        collection_name: str
    ) -> Dict[str, Any]:
        """
        Get detailed information about a vector collection.
        """

        vdb = DatabaseConnection.get_vector()

        try:
            collection = vdb.get_collection(collection_name)
            if collection is None:
                return {
                    "name": collection_name,
                    "exists": False,
                    "error": "Collection not found"
                }
            return {
                "name": collection_name,
                "exists": True,
                "document_count": vdb.count(collection_name),
                "metadata": collection.get('metadata', {})
            }
        except Exception as error:
            return {
                "name": collection_name,
                "exists": False,
                "error": str(error)
            }
        finally:
            vdb.close()

    def get_documents(
        self,
        collection_name: str,
        document_ids: Optional[List[str]] = None,
        metadata_filter: Optional[Dict] = None,
        limit: int = 100,
        offset: int = 0
    ) -> Dict[str, Any]:
        """
        Get documents from a vector collection.
        """

        vdb = DatabaseConnection.get_vector()

        try:
            results = vdb.get(
                collection_name,
                ids=document_ids,
                where=metadata_filter,
                include=["documents", "metadatas"]
            )
            return self._paginate(results, limit, offset)
        except Exception as error:
            return self._empty_result(error)
        finally:
            vdb.close()

    def search(
        self,
        collection_name: str,
        query_text: str,
        n_results: int = 10,
        metadata_filter: Optional[Dict] = None
    ) -> Dict[str, Any]:
        """
        Search for similar documents using vector similarity.
        """

        vdb = DatabaseConnection.get_vector()

        try:
            results = vdb.query(
                collection_name,
                query_texts=[query_text],
                n_results=n_results,
                where=metadata_filter,
                include=["documents", "metadatas", "distances"]
            )
            return self._format_results(results, query_text, n_results)
        except Exception as error:
            return {
                "results": [],
                "query": query_text,
                "error": str(error)
            }
        finally:
            vdb.close()

    def _paginate(
        self,
        results: Dict,
        limit: int,
        offset: int
    ) -> Dict[str, Any]:
        """
        Apply pagination to vector results.
        """

        total = len(results.get("ids", []))
        start = offset
        end = offset + limit if limit else total
        return {
            "ids": results["ids"][start:end],
            "documents": results.get("documents", [])[start:end],
            "metadatas": results.get("metadatas", [])[start:end],
            "total": total,
            "limit": limit,
            "offset": offset,
            "has_more": end < total
        }

    def _format_results(
        self,
        results: Dict,
        query_text: str,
        n_results: int
    ) -> Dict[str, Any]:
        """
        Format vector search results.
        """

        ids = results.get("ids", [[]])[0]
        documents = results.get("documents", [[]])[0]
        metadatas = results.get("metadatas", [[]])[0] if results.get("metadatas") else [None] * len(ids)
        distances = results.get("distances", [[]])[0] if results.get("distances") else [None] * len(ids)

        formatted = [
            {
                "id": ids[i],
                "document": documents[i],
                "metadata": metadatas[i],
                "similarity_score": distances[i]
            }
            for i in range(len(ids))
        ]
        return {
            "results": formatted,
            "query": query_text,
            "n_results": n_results
        }

    def _empty_result(
        self,
        error: Exception
    ) -> Dict[str, Any]:
        """
        Return empty result with error.
        """

        return {
            "ids": [],
            "documents": [],
            "metadatas": [],
            "total": 0,
            "error": str(error)
        }

class StatisticsAPI:
    """
    API for database statistics.
    """

    def get_all(self) -> Dict[str, Any]:
        """
        Get comprehensive statistics for all databases.
        """

        stats = {
            "timestamp": time.time_ns(),
            "tables": {"count": 0, "total_rows": 0},
            "collections": {"count": 0, "total_documents": 0}
        }

        self._collect_tables(stats)
        self._collect_collections(stats)
        return stats

    def _collect_tables(
        self,
        stats: Dict[str, Any]
    ) -> None:
        """
        Collect table statistics.
        """

        try:
            db = DatabaseConnection.get_main()
            table_names = db.get_all_tables_names()

            stats["tables"]["count"] = len(table_names)
            stats["tables"]["total_rows"] = sum(
                len(db.get_all_data(table))
                for table in table_names
            )

            db.close()
        except Exception as error:
            stats["tables"]["error"] = str(error)

    def _collect_collections(
        self,
        stats: Dict[str, Any]
    ) -> None:
        """
        Collect collection statistics.
        """

        vdb = None
        try:
            vdb = DatabaseConnection.get_vector()
            collections = vdb.list_collections()

            stats["collections"]["count"] = len(collections)
            stats["collections"]["total_documents"] = sum(
                vdb.count(coll['name'])
                for coll in collections
            )
        except Exception as error:
            stats["collections"]["error"] = str(error)
        finally:
            if vdb is not None:
                try:
                    vdb.close()
                except Exception:
                    # suppress close errors to avoid masking original exceptions
                    pass

class DashboardAPI:
    """
    Main Dashboard API class providing access to all monitoring operations.

    Organizes functionality into logical groups:
        health: System health monitoring
        tables: Table operations
        vector: Vector collection operations
        statistics: Database-wide statistics

    Example:
        api = DashboardAPI()

        # Check health
        health = api.health.check()

        # List tables
        tables = api.tables.list_all()

        # Get table data
        data = api.tables.get_data("users", limit=50)

        # List collections
        collections = api.vector.list_all()

        # Search vectors
        results = api.vector.search("docs", "machine learning")

        # Get statistics
        stats = api.statistics.get_all()
    """

    def __init__(self):
        """
        Initialize Dashboard API with all sub-APIs.
        """

        self.health = HealthAPI()
        self.tables = TableAPI()
        self.vector = VectorAPI()
        self.statistics = StatisticsAPI()
        self.links = LinksAPI()

    def get_summary(self) -> Dict[str, Any]:
        """
        Get quick summary of entire database system.

        Returns:
            Dictionary with health status and key metrics
        """

        health = self.health.check()
        stats = self.statistics.get_all()
        return {
            "status": health["status"],
            "timestamp": health["timestamp"],
            "summary": {
                "tables": stats["tables"],
                "collections": stats["collections"]
            },
            "health_details": health["databases"],
            "database_links": self.links.list_all()
        }



"""
Module containing the SysAdd class, which is used to add data to a table in the database.
"""

from typing import List
from skypydb.database.reactive_db import ReactiveDatabase

class SysAdd:
    def __init__(
        self,
        db: "ReactiveDatabase",
        table_name: str
    ):
        self.db = db
        self.table_name = table_name

    def add(
        self,
        **kwargs
    ) -> List[str]:
        """
        Add data to the table.

        Each keyword argument can be a single value or a list.
        If a value is a list, multiple rows will be inserted.
        IDs and timestamps are automatically generated.

        Args:
            **kwargs: Column names and values (can be lists)

        Returns:
            List of IDs for inserted rows

        Example:
            table.add(
                title=["doc1", "doc2"],
                user_id=["user123"],
                content=["content1", "content2"]
            )
        """

        # handle "auto" for id field
        if 'id' in kwargs:
            if kwargs['id'] == ['auto'] or kwargs['id'] == 'auto':
                del kwargs['id']

        # determine number of rows to insert
        max_length = 1
        for key, value in kwargs.items():
            if isinstance(value, list):
                if not value:
                    raise ValueError(f"Empty list provided for '{key}'")
                max_length = max(max_length, len(value))

        # prepare data for each row
        inserted_ids = []
        for row_index in range(max_length):
            row_data = {}
            for key, value in kwargs.items():
                if isinstance(value, list):
                    row_data[key] = value[row_index] if row_index < len(value) else value[-1]
                else:
                    row_data[key] = value

            # validate data against table config
            validated_data = self.db.validate_data_with_config(self.table_name, row_data)

            # insert row with validated data
            row_id = self.db.add_data(self.table_name, validated_data, generate_id=True)

            inserted_ids.append(row_id)
        return inserted_ids



"""
Module containing the SysPassword class, which is used to securely manage passwords.
"""

import base64
import secrets
from typing import Optional
from skypydb.errors import EncryptionError
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend

class SysPassword:
    def __init__(
        self,
        iterations: int = 100000
    ):
        self.iterations = iterations

    def _derive_key(
        self,
        password: str,
        salt: Optional[bytes] = None
    ) -> bytes:
        """
        Derive a 256-bit encryption key from a password using PBKDF2HMAC.

        Args:
            password: Master password/key
            salt: Required, non-empty salt for PBKDF2HMAC

        Returns:
            32-byte encryption key
        """

        if not salt:
            raise EncryptionError("Encryption salt must be provided and non-empty.")

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,  # 256 bits
            salt=salt,
            iterations=self.iterations,
            backend=default_backend()
        )
        return kdf.derive(password.encode('utf-8'))

    def hash_password(
        self,
        password: str
    ) -> str:
        """
        Create a secure hash of a password for storage using PBKDF2HMAC with a random salt.

        Args:
            password: Password to hash

        Returns:
            Base64-encoded hash with format: salt|hash
        """

        # generate a random salt
        salt = secrets.token_bytes(32)

        # hash the password
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=self.iterations,
            backend=default_backend()
        )

        password_hash = kdf.derive(password.encode('utf-8'))

        # combine salt and hash
        combined = salt + password_hash
        return base64.b64encode(combined).decode('utf-8')

    def verify_password(
        self,
        password: str,
        stored_hash: str
    ) -> bool:
        """
        Verify a password against a stored hash.

        Args:
            password: Password to verify
            stored_hash: Stored hash from hash_password()

        Returns:
            True if password matches, False otherwise
        """

        try:
            # decode the stored hash
            combined = base64.b64decode(stored_hash.encode('utf-8'))

            # extract salt and hash
            salt = combined[:32]
            stored_password_hash = combined[32:]

            # hash the provided password with the same salt
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=self.iterations,
                backend=default_backend()
            )

            password_hash = kdf.derive(password.encode('utf-8'))

            # compare hashes using constant-time comparison
            return secrets.compare_digest(password_hash, stored_password_hash)
        except Exception:
            return False



"""
Custom exceptions for Skypydb.
"""

class SkypydbError(Exception):
    """
    Base exception for all Skypydb errors.
    """

    CODE = "SKY001"
    default_message = "An unexpected error occurred in Skypydb."

    # initialize the SkypydbError instance for handling and formatting error messages
    def __init__(
        self,
        message=None,
    ):
        """
        Initialize the SkypydbError instance.

        Args:
            message (str, optional): The error message. Defaults to None.
        """

        self.message = message

        if self.message:
            formatted_message = f"[{self.CODE}] {self.message}"
        else:
            # use a class-specific default message when provided, otherwise fall back
            # to a generic, user-friendly message when no default_message is defined.
            default_msg = getattr(
                self,
                "default_message",
                "An unexpected error occurred.",
            )
            formatted_message = f"[{self.CODE}] {default_msg}"

        super().__init__(formatted_message)

class TableNotFoundError(SkypydbError):
    """
    Raised when a table is not found.
    """

    CODE = "SKY101"
    default_message = "Table not found."

class TableAlreadyExistsError(SkypydbError):
    """
    Raised when trying to create a table that already exists.
    """

    CODE = "SKY102"
    default_message = "Table already exists."

class DatabaseError(SkypydbError):
    """
    Raised when a database-level operation fails.
    """

    CODE = "SKY103"
    default_message = (
        "Database operation failed. This may indicate a connectivity issue, "
        "invalid query, or transaction problem. Check database logs and configuration."
    )

class InvalidSearchError(SkypydbError):
    """
    Raised when search parameters are invalid.
    """

    CODE = "SKY201"
    default_message = (
        "One or more search parameters are invalid. "
        "Check parameter names, types, and value ranges for your search query."
    )

class SecurityError(SkypydbError):
    """
    Raised when a security operation fails.
    """

    CODE = "SKY301"
    default_message = (
        "Security operation failed. Possible authentication or authorization issue; "
        "see logs for details."
    )

class ValidationError(SkypydbError):
    """
    Raised when input validation fails.
    """

    CODE = "SKY302"
    default_message = "Input validation failed."

class EncryptionError(SkypydbError):
    """
    Raised when encryption/decryption operations fail.
    """

    CODE = "SKY303"
    default_message = "Encryption or decryption operation failed."

class CollectionNotFoundError(SkypydbError):
    """
    Raised when a vector collection is not found.
    """

    CODE = "SKY401"
    default_message = "Collection not found."

class CollectionAlreadyExistsError(SkypydbError):
    """
    Raised when trying to create a collection that already exists.
    """

    CODE = "SKY402"
    default_message = "Collection already exists."

class EmbeddingError(SkypydbError):
    """
    Raised when embedding generation fails.
    """

    CODE = "SKY403"
    default_message = "Embedding generation failed."

class VectorSearchError(SkypydbError):
    """
    Raised when vector similarity search fails.
    """

    CODE = "SKY404"
    default_message = "Vector similarity search failed."



"""
Vector Client API for Skypydb.
"""

import os
from typing import (
    Any,
    Dict,
    Optional
)
from skypydb.database.vector_db import VectorDatabase
from skypydb.embeddings import get_embedding_function
from skypydb.api.collection import Collection
from skypydb.database.database_linker import DatabaseLinker
from skypydb.api.mixins.vector import (
    SysCreate,
    SysGet,
    SysList,
    SysDelete,
    Utils
)

class VectorClient(
    SysCreate,
    SysGet,
    SysList,
    SysDelete,
    Utils
):
    """
    Vector client for interacting with Skypydb.
    """

    def __init__(
        self,
        path: str = "./db/_generated/vector.db",
        embedding_provider: str = "ollama",
        embedding_model_config: Optional[Dict[str, Any]] = None
    ):
        """
        Initialize Vector Client.

        Args:
            path: Path to the database file. Defaults to ./db/_generated/vector.db
            embedding_provider: Embedding provider (ollama, openai, sentence-transformers)
            embedding_model_config: Provider-specific config dictionary.

        Example:
            # Basic usage with defaults
            client = skypydb.VectorClient()

            # With explicit provider config
            client = skypydb.VectorClient(
                embedding_provider="openai",
                embedding_model_config={
                    "api_key": "your-openai-api-key",
                    "model": "text-embedding-3-small"
                }
            )
        """

        # constant to define the path to the database file
        DB_PATH = path

        db_dir = os.path.dirname(DB_PATH)
        if db_dir and not os.path.exists(db_dir):
            os.makedirs(db_dir, exist_ok=True)

        self.path = DB_PATH
        if embedding_model_config is not None and not isinstance(embedding_model_config, dict):
            raise TypeError("embedding_model_config must be a dictionary when provided.")

        provider = embedding_provider.lower().strip().replace("_", "-")
        model_config: Dict[str, Any] = dict(embedding_model_config or {})

        # set up embedding function
        self._embedding_function = get_embedding_function(
            provider=provider,
            **model_config
        )

        self.database_linker = DatabaseLinker()

        # initialize vector database
        self._db = VectorDatabase(
            path=DB_PATH,
            embedding_function=self._embedding_function
        )
        self.database_linker.ensure_db_link_metadata(DB_PATH, db_type="vector")

        # cache for collection instances
        self._collections: Dict[str, Collection] = {}

    def close(
        self,
    ) -> None:
        """
        Close the database connection.
        """

        self._db.close()
        self._collections.clear()



"""
Sentence Transformers embedding functions for vector operations.
"""

from typing import (
    Any,
    List,
    Optional
)
from skypydb.embeddings.mixins import (
    EmbeddingsFn,
    Utils
)

class SentenceTransformerEmbedding(
    EmbeddingsFn,
    Utils
):
    def __init__(
        self,
        model: str = "all-MiniLM-L6-v2",
        device: Optional[str] = None,
        normalize_embeddings: bool = False,
        dimension: Optional[int] = None
    ):
        """
        Initialize Sentence Transformers embedding function.

        Args:
            model: Sentence Transformers model name.
            device: Optional device override (e.g. cpu, cuda).
            normalize_embeddings: Whether to return normalized embeddings.
        """

        super().__init__(dimension=dimension)
        self.model = model
        self.device = device
        self.normalize_embeddings = normalize_embeddings

        try:
            from sentence_transformers import SentenceTransformer
        except ImportError as exc:
            raise ImportError(
                "Sentence Transformers embedding provider requires the "
                "`sentence-transformers` package. Install it with "
                "`pip install sentence-transformers`."
            ) from exc
        model_kwargs = {}
        if self.device:
            model_kwargs["device"] = self.device
        self._model = SentenceTransformer(model_name_or_path=self.model, **model_kwargs)

    @staticmethod
    def _to_list(vector: Any) -> List[float]:
        """
        Convert a vector to a list of floats.
        """

        if hasattr(vector, "tolist"):
            return list(vector.tolist())
        return list(vector)

    def embed(
        self,
        texts: List[str]
    ) -> List[List[float]]:
        """
        Generate embeddings for a list of texts using Sentence Transformers.
        """

        if not texts:
            return []

        vectors = self._model.encode(
            texts,
            convert_to_numpy=False,
            normalize_embeddings=self.normalize_embeddings
        )
        embeddings = [self._to_list(vector) for vector in vectors]
        if self._dimension is None and embeddings:
            self._dimension = len(embeddings[0])
        return embeddings


"""
Module containing the SysGenerator class, which is used to generate secure random encryption key and salt.
"""

import secrets
from skypydb.errors import EncryptionError

class SysGenerator:
    @staticmethod
    def generate_key() -> str:
        """
        Generate a secure random encryption key.

        Returns:
            Random 256-bit key encoded as hex string
        """

        return secrets.token_hex(32)  # 32 bytes = 256 bits

    @staticmethod
    def generate_salt(length: int = 32) -> bytes:
        """
        Generate a secure random salt.

        Args:
            length: Salt length in bytes (default: 32)

        Returns:
            Random salt as bytes
        """

        if length <= 0:
            raise EncryptionError("Salt length must be positive.")
        return secrets.token_bytes(length)



"""
Vector database operations module.
"""

from skypydb.database.mixins.vector.utils import cosine_similarity, euclidean_distance
from skypydb.database.mixins.vector.sysembeddings import SysEmbeddings
from skypydb.database.mixins.vector.sysadd import SysAdd
from skypydb.database.mixins.vector.sysupdate import SysUpdate
from skypydb.database.mixins.vector.sysquery import SysQuery
from skypydb.database.mixins.vector.vsysget import VSysGet
from skypydb.database.mixins.vector.vsysdelete import VSysDelete
from skypydb.database.mixins.vector.collections import (
    AuditCollections,
    SysCreate,
    SysGet,
    SysCount,
    SysDelete
)

__all__ = [
    cosine_similarity,
    euclidean_distance,
    SysEmbeddings,
    SysAdd,
    SysUpdate,
    SysQuery,
    VSysGet,
    VSysDelete,
    AuditCollections,
    SysCreate,
    SysGet,
    SysCount,
    SysDelete
]



"""
Module containing the SysUpdate class, which is used to update items in the collection.
"""

import json
from typing import (
    Dict,
    List,
    Optional,
    Any
)
from skypydb.security.validation import InputValidator

class SysUpdate:
    def update(
        self,
        collection_name: str,
        ids: List[str],
        embeddings: Optional[List[List[float]]] = None,
        documents: Optional[List[str]] = None,
        metadatas: Optional[List[Dict[str, Any]]] = None
    ) -> None:
        """
        Update items in a collection.

        Args:
            collection_name: Name of the collection
            ids: List of IDs to update
            embeddings: Optional new embeddings
            documents: Optional new documents (will be embedded)
            metadatas: Optional new metadata
        """

        collection_name = InputValidator.validate_table_name(collection_name)
        if not self.collection_exists(collection_name):
            raise ValueError(f"Collection '{collection_name}' not found")
        if embeddings is None and documents is not None:
            if self.embedding_function is None:
                raise ValueError(
                    "Documents provided but no embedding function set."
                )
            embeddings = self.embedding_function(documents)

        cursor = self.conn.cursor()

        for i, item_id in enumerate(ids):
            updates = []
            params = []
            if embeddings is not None:
                updates.append("embedding = ?")
                params.append(json.dumps(embeddings[i]))
            if documents is not None:
                updates.append("document = ?")
                params.append(documents[i])
            if metadatas is not None:
                updates.append("metadata = ?")
                params.append(json.dumps(metadatas[i]) if metadatas[i] else None)
            if updates:
                params.append(item_id)
                cursor.execute(
                    f"UPDATE [vec_{collection_name}] SET {', '.join(updates)} WHERE id = ?",
                    params
                )
        self.conn.commit()


"""
Module containing the VSysGet class, which is used to get items in the collection.
"""

import json
from typing import (
    Any,
    Dict,
    List,
    Optional
)
from skypydb.security.validation import InputValidator

class VSysGet:
    def get(
        self,
        collection_name: str,
        ids: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, str]] = None,
        include: Optional[List[str]] = None
    ) -> Dict[str, List[Any]]:
        """
        Get items from a collection by ID or filter.
        
        Args:
            collection_name: Name of the collection
            ids: Optional list of IDs to retrieve
            where: Optional metadata filter
            where_document: Optional document filter
            include: Optional list of fields to include (embeddings, documents, metadatas)
            
        Returns:
            Dictionary with lists of ids, embeddings, documents, and metadatas
        """

        collection_name = InputValidator.validate_table_name(collection_name)
        if not self.collection_exists(collection_name):
            raise ValueError(f"Collection '{collection_name}' not found")

        include = include or ["embeddings", "documents", "metadatas"]

        cursor = self.conn.cursor()

        if ids is not None:
            placeholders = ", ".join(["?" for _ in ids])
            cursor.execute(
                f"SELECT * FROM [vec_{collection_name}] WHERE id IN ({placeholders})",
                list(ids)
            )
        else:
            cursor.execute(f"SELECT * FROM [vec_{collection_name}]")

        results = {
            "ids": [],
            "embeddings": [] if "embeddings" in include else None,
            "documents": [] if "documents" in include else None,
            "metadatas": [] if "metadatas" in include else None,
        }
        for row in cursor.fetchall():
            item = {
                "id": row["id"],
                "document": row["document"],
                "embedding": json.loads(row["embedding"]),
                "metadata": json.loads(row["metadata"]) if row["metadata"] else None,
            }
            # apply filters
            if not self._matches_filters(item, where, where_document):
                continue

            results["ids"].append(item["id"])
            if results["embeddings"] is not None:
                results["embeddings"].append(item["embedding"])
            if results["documents"] is not None:
                results["documents"].append(item["document"])
            if results["metadatas"] is not None:
                results["metadatas"].append(item["metadata"])
        return results

    def _get_all_items(
        self,
        collection_name: str
    ) -> List[Dict[str, Any]]:
        """
        Get all items from a collection.
        """

        cursor = self.conn.cursor()
        
        cursor.execute(f"SELECT * FROM [vec_{collection_name}]")

        items = []
        for row in cursor.fetchall():
            items.append({
                "id": row["id"],
                "document": row["document"],
                "embedding": json.loads(row["embedding"]),
                "metadata": json.loads(row["metadata"]) if row["metadata"] else None,
                "created_at": row["created_at"]
            })
        return items


"""
Module containing the SysList class, which is used to list all collections in the database.
"""

from typing import (
    List,
    TYPE_CHECKING
)

if TYPE_CHECKING:
    from skypydb.api.collection import Collection

class SysList:
    def list_collections(
        self
    ) -> List["Collection"]:
        """
        List all collections in the database.

        Returns:
            List of Collection instances

        Example:
            for collection in client.list_collections():
                print(f"Collection: {collection.name}")
                print(f"Documents: {collection.count()}")
        """

        collections = []

        for collection_info in self._db.list_collections():
            name = collection_info["name"]
            # use cached instance if available
            if name in self._collections:
                collections.append(self._collections[name])
            else:
                from skypydb.api.collection import Collection
                collection = Collection(
                    db=self._db,
                    name=name,
                    metadata=collection_info.get("metadata")
                )
                self._collections[name] = collection
                collections.append(collection)
        return collections


"""
Module containing the SysCreate class, which is used to create a table.
"""

from typing import Dict
from skypydb.errors import TableAlreadyExistsError
from skypydb.table.table import Table
from skypydb.api.mixins.reactive.schema_loader import load_schema

class SysCreate:
    def create_table(self) -> Dict[str, Table]:
        """
        Create all tables defined in db/schema.py.

        Returns:
            Dictionary mapping table names to Table instances

        Raises:
            TableAlreadyExistsError: If any table already exists
            ValueError: If schema file is missing or invalid

        Example:
            # Define your schema in db/schema.py, then:
            client = skypydb.Client()
            tables = client.create_table()

            # Access tables
            users_table = tables["users"]
            posts_table = tables["posts"]
        """

        schema = load_schema()

        # create all tables from schema
        created_tables: Dict[str, Table] = {}
        table_names = schema.get_all_table_names()
        existing = [name for name in table_names if self.db.table_exists(name)]
        if existing:
            raise TableAlreadyExistsError(
                f"Tables already exist in the database: {', '.join(existing)}"
            )
        for table_name in table_names:
            table_def = schema.get_table_definition(table_name)
            if table_def is None:
                continue
            # create table with schema definition
            self.db.create_table(table_name, table_def)
            created_tables[table_name] = Table(self.db, table_name)
        return created_tables


"""
OpenAI embedding functions for vector operations.
"""

import os
from typing import (
    List,
    Optional,
    Any
)
from skypydb.embeddings.mixins import (
    EmbeddingsFn,
    Utils
)

class OpenAIEmbedding(
    EmbeddingsFn,
    Utils
):
    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = "text-embedding-3-small",
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        project: Optional[str] = None,
        timeout: Optional[float] = None,
        dimension: Optional[int] = None
    ):
        """
        Initialize OpenAI embedding function.

        Args:
            api_key: OpenAI API key. If not provided, OPENAI_API_KEY is used.
            model: OpenAI embedding model.
            base_url: Optional custom OpenAI-compatible base URL.
            organization: Optional OpenAI organization ID.
            project: Optional OpenAI project ID.
            timeout: Optional timeout in seconds.
        """

        super().__init__(dimension=dimension)
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        if not self.api_key:
            raise ValueError(
                "OpenAI API key is required. Provide `api_key` or set OPENAI_API_KEY."
            )

        self.model = model
        self.base_url = base_url
        self.organization = organization
        self.project = project
        self.timeout = timeout

        try:
            from openai import OpenAI
        except ImportError as exc:
            raise ImportError(
                "OpenAI embedding provider requires the `openai` package. "
                "Install it with `pip install openai`."
            ) from exc

        client_kwargs: dict[str, Any] = {
            "api_key": self.api_key
        }
        if self.base_url is not None:
            client_kwargs["base_url"] = self.base_url
        if self.organization is not None:
            client_kwargs["organization"] = self.organization
        if self.project is not None:
            client_kwargs["project"] = self.project
        if self.timeout is not None:
            client_kwargs["timeout"] = self.timeout

        self._client = OpenAI(**client_kwargs)

    def embed(
        self,
        texts: List[str]
    ) -> List[List[float]]:
        """
        Generate embeddings for a list of texts using OpenAI API.
        """

        if not texts:
            return []

        response = self._client.embeddings.create(
            model=self.model,
            input=texts
        )
        embeddings = [list(item.embedding) for item in response.data]
        if self._dimension is None and embeddings:
            self._dimension = len(embeddings[0])
        return embeddings


"""
Module containing the SysDef class, which is used to define table in the database.
"""

from typing import (
    List,
    Dict,
    Any,
    Optional,
    TYPE_CHECKING
)
from skypydb.schema.values import Validator
from skypydb.schema.mixins.schema.sysschema import SysSchema

if TYPE_CHECKING:
    from skypydb.schema.schema import TableDefinition

class SysDef:
    def __init__(
        self,
        columns: Dict[str, Validator],
        table_name: Optional[str] = None
    ):
        self.columns = columns
        self.indexes: List[Dict[str, Any]] = []
        self.table_name = table_name

def defineTable(
    columns: Dict[str, Validator]
) -> "TableDefinition":
    """
    Define a table with its columns and types.

    Args:
        columns: Dictionary mapping column names to validators

    Returns:
        TableDefinition that can be configured with indexes

    Example:
        users = defineTable({
            "name": value.string(),
            "email": value.string(),
            "age": value.int64(),
            "active": value.boolean(),
            "bio": value.optional(value.string())
        })
        .index("by_email", ["email"])
        .index("by_age", ["age"])
    """

    from skypydb.schema.schema import TableDefinition
    return TableDefinition(columns)

def defineSchema(
    tables: Dict[str, "TableDefinition"]
) -> "SysSchema":
    """
    Define a schema with multiple tables.

    Args:
        tables: Dictionary mapping table names to table definitions

    Returns:
        Schema object containing all tables

    Example:
        schema = defineSchema({
            "users": defineTable({
                "name": value.string(),
                "email": value.string()
            })
            .index("by_name", ["name"]),
            
            "posts": defineTable({
                "title": value.string(),
                "content": value.string()
            })
            .index("by_title", ["title"])
        })
    """

    # set table names in definitions
    for table_name, table_def in tables.items():
        table_def.table_name = table_name
    return SysSchema(tables)



"""
Module containing the RSysDelete class, which is used to delete data from a table.
"""

import sqlite3
from typing import Optional
from skypydb.security.validation import InputValidator
from skypydb.errors import TableNotFoundError
from skypydb.database.mixins.reactive.tables.audit import AuditTable

class RSysDelete:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

        self.audit = AuditTable(conn=self.conn)

    def delete(
        self,
        table_name: str,
        **filters
    ) -> int:
        """
        Delete data from a table based on filters.

        Args:
            table_name: Name of the table
            **filters: Filters as keyword arguments (column name = value)

        Returns:
            Number of rows deleted

        Example:
            db.delete(
                table_name="my_table",
                id="123"
            )
            db.delete(
                table_name="my_table",
                user_id="user123",
                title="document"
            )

        Raises:
            ValidationError: If input parameters are invalid
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        # validate filters
        if filters:
            filters = InputValidator.validate_filter_dict(filters)
        if not self.audit.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")
        if not filters:
            # safety check; don't allow deleting all rows without explicit filters
            raise ValueError("Cannot delete without filters. Use filters to specify which rows to delete.")

        conditions = []
        params = []

        # build WHERE clause from filters
        for column, value in filters.items():
            # handle list values, use IN clause
            if isinstance(value, list) and len(value) > 0:
                placeholders = ", ".join(["?" for _ in value])
                conditions.append(f"[{column}] IN ({placeholders})")
                params.extend([str(v) for v in value])
            else:
                conditions.append(f"[{column}] = ?")
                params.append(str(value))

        # build DELETE query
        where_clause = " AND ".join(conditions)
        query = f"DELETE FROM [{table_name}] WHERE {where_clause}"

        cursor = self.conn.cursor()

        cursor.execute(query, params)
        self.conn.commit()
        return cursor.rowcount



"""
Skypydb API Server
"""

import os
from typing import (
    Dict,
    Any,
    Optional
)
from fastapi import (
    FastAPI,
    HTTPException,
    Header
)
from fastapi.middleware.cors import CORSMiddleware
import uvicorn
from skypydb.server.dashboard_server import DashboardAPI

app = FastAPI(
    title="SkypyDB Dashboard API",
    description="REST API for monitoring SkypyDB databases",
    version="1.0.0"
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:3000"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
dashboard_api = DashboardAPI()

def update_db_paths(
    main_path: Optional[str],
    vector_path: Optional[str]
):
    """
    Update database paths from request headers.
    """

    if main_path:
        os.environ['SKYPYDB_PATH'] = main_path
    if vector_path:
        os.environ['SKYPYDB_VECTOR_PATH'] = vector_path

@app.get("/api/health")
async def health_check(
    x_skypydb_path: Optional[str] = Header(None),
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Check health status of all database components.
    """

    try:
        update_db_paths(x_skypydb_path, x_skypydb_vector_path)
        return dashboard_api.health.check()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/summary")
async def get_summary(
    x_skypydb_path: Optional[str] = Header(None),
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Get quick summary of entire database system.
    """

    try:
        update_db_paths(x_skypydb_path, x_skypydb_vector_path)
        return dashboard_api.get_summary()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/statistics")
async def get_statistics(
    x_skypydb_path: Optional[str] = Header(None),
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Get comprehensive statistics for all databases.
    """

    try:
        update_db_paths(x_skypydb_path, x_skypydb_vector_path)
        return dashboard_api.statistics.get_all()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/databaselinks")
async def get_database_links():
    """
    Get discovered database type/path links from project root.
    """

    try:
        return dashboard_api.links.list_all()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/tables")
async def list_tables(
    x_skypydb_path: Optional[str] = Header(None)
):
    """
    Get all tables with metadata and row counts.
    """

    try:
        if x_skypydb_path:
            os.environ['SKYPYDB_PATH'] = x_skypydb_path
        return dashboard_api.tables.list_all()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/tables/{table_name}/schema")
async def get_table_schema(
    table_name: str,
    x_skypydb_path: Optional[str] = Header(None)
):
    """
    Get schema information for a table.
    """

    try:
        if x_skypydb_path:
            os.environ['SKYPYDB_PATH'] = x_skypydb_path
        return dashboard_api.tables.get_schema(table_name)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/tables/{table_name}/data")
async def get_table_data(
    table_name: str,
    limit: int = 100,
    offset: int = 0,
    x_skypydb_path: Optional[str] = Header(None)
):
    """
    Get paginated data from a table.
    """

    try:
        if x_skypydb_path:
            os.environ['SKYPYDB_PATH'] = x_skypydb_path
        return dashboard_api.tables.get_data(table_name, limit=limit, offset=offset)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/tables/{table_name}/search")
async def search_table(
    table_name: str,
    query: Optional[str] = None,
    limit: int = 100,
    x_skypydb_path: Optional[str] = Header(None)
):
    """
    Search table data with filters.
    """

    try:
        if x_skypydb_path:
            os.environ['SKYPYDB_PATH'] = x_skypydb_path
        return dashboard_api.tables.search(table_name, query=query, limit=limit)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/api/collections")
async def list_collections(
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Get all vector collections with document counts.
    """

    try:
        if x_skypydb_vector_path:
            os.environ['SKYPYDB_VECTOR_PATH'] = x_skypydb_vector_path
        return dashboard_api.vector.list_all()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.get("/api/collections/{collection_name}")
async def get_collection_details(
    collection_name: str,
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Get detailed information about a vector collection.
    """

    try:
        if x_skypydb_vector_path:
            os.environ['SKYPYDB_VECTOR_PATH'] = x_skypydb_vector_path
        return dashboard_api.vector.get_details(collection_name)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/api/collections/{collection_name}/documents")
async def get_collection_documents(
    collection_name: str,
    body: Dict[str, Any],
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Get documents from a vector collection.
    """

    try:
        if x_skypydb_vector_path:
            os.environ['SKYPYDB_VECTOR_PATH'] = x_skypydb_vector_path

        limit = body.get('limit', 100)
        offset = body.get('offset', 0)
        document_ids = body.get('document_ids')
        metadata_filter = body.get('metadata_filter')
        return dashboard_api.vector.get_documents(
            collection_name,
            document_ids=document_ids,
            metadata_filter=metadata_filter,
            limit=limit,
            offset=offset
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.post("/api/collections/{collection_name}/search")
async def search_vectors(
    collection_name: str,
    body: Dict[str, Any],
    x_skypydb_vector_path: Optional[str] = Header(None)
):
    """
    Search for similar documents using vector similarity.
    """

    try:
        if x_skypydb_vector_path:
            os.environ['SKYPYDB_VECTOR_PATH'] = x_skypydb_vector_path

        query_text = body.get('query_text', '')
        n_results = body.get('n_results', 10)
        metadata_filter = body.get('metadata_filter')
        return dashboard_api.vector.search(
            collection_name,
            query_text,
            n_results=n_results,
            metadata_filter=metadata_filter
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

if __name__ == "__main__":
    print("Starting SkypyDB API Server.")
    print("API will be available at: http://localhost:8000/api")
    print("\nEndpoints:")
    print("  - GET  /api/health")
    print("  - GET  /api/summary")
    print("  - GET  /api/statistics")
    print("  - GET  /api/databaselinks")
    print("  - GET  /api/tables")
    print("  - GET  /api/tables/{name}/schema")
    print("  - GET  /api/tables/{name}/data")
    print("  - GET  /api/collections")
    print("  - POST /api/collections/{name}/search")
    print("\nPress Ctrl+C to stop")
    
    uvicorn.run(
        app,
        host="0.0.0.0",
        port=8000
    )



"""
Module containing the SysSanitize class, which is used to sanitize data in the database.
"""

from typing import Any

class SysSanitize:
    @staticmethod
    def sanitize_string(
        value: str
    ) -> str:
        """
        Sanitize a string value by removing potentially dangerous characters.

        Args:
            value: String to sanitize

        Returns:
            Sanitized string
        """

        if not isinstance(value, str):
            return str(value)

        # remove null bytes
        value = value.replace('\x00', '')

        # note: We don't strip SQL characters here because data should be
        # parameterized in queries, but we do basic sanitization
        return value

def sanitize_input(
    self,
    value: Any
) -> Any:
    """
    Convenience function to sanitize an input value.

    Args:
        value: Value to sanitize
        sys_sanitize: SysSanitize instance to use for sanitization

    Returns:
        Sanitized value
    """

    if isinstance(value, str):
        return self.sanitize_string(value)
    return value



"""
Vector collections module.
"""

from skypydb.database.mixins.vector.collections.audit import AuditCollections
from skypydb.database.mixins.vector.collections.syscreate import SysCreate
from skypydb.database.mixins.vector.collections.sysget import SysGet
from skypydb.database.mixins.vector.collections.syscount import SysCount
from skypydb.database.mixins.vector.collections.sysdelete import SysDelete

__all__ = [
    AuditCollections,
    SysCreate,
    SysGet,
    SysCount,
    SysDelete
]



"""
Table mixins module for Skypydb.
"""

from skypydb.table.mixins.sysadd import SysAdd
from skypydb.table.mixins.sysdelete import SysDelete
from skypydb.table.mixins.sysget import SysGet
from skypydb.table.mixins.syssearch import SysSearch

__all__ = [
    "SysAdd",
    "SysDelete",
    "SysGet",
    "SysSearch"
]



"""
Collection class for managing vector collections.
"""

from typing import (
    Any,
    Dict,
    Optional,
    TYPE_CHECKING
)
from skypydb.api.mixins.vector.collection import (
    SysAdd,
    SysGet,
    SysQuery,
    SysUpdate,
    SysDelete,
    Utils
)

if TYPE_CHECKING:
    from skypydb.database.vector_db import VectorDatabase

class Collection(
    SysAdd,
    SysGet,
    SysQuery,
    SysUpdate,
    SysDelete,
    Utils
):
    """
    Represents a vector collection in the database.
    """

    def __init__(
        self,
        db: "VectorDatabase",
        name: str,
        metadata: Optional[Dict[str, Any]] = None
    ):
        """
        Initialize collection.

        Args:
            db: VectorDatabase instance
            name: Name of the collection
            metadata: Optional collection metadata
        """

        self._db = db
        self._name = name
        self._metadata = metadata or {}


"""
Module containing the SysCheck class, which is used to check SQL queries for potential SQL injection patterns.
"""

import re
from skypydb.security.constants import SQL_INJECTION_PATTERNS

class SysCheck:
    @classmethod
    def _contains_sql_injection(
        cls,
        value: str
    ) -> bool:
        """
        Check if a value contains potential SQL injection patterns.

        Args:
            value: String to check

        Returns:
            True if potentially dangerous patterns detected
        """

        value_upper = value.upper()

        for pattern in SQL_INJECTION_PATTERNS:
            if re.search(pattern, value_upper, re.IGNORECASE):
                return True
        return False



"""
Module containing the SysDelete class, which is used to delete a collection.
"""

class SysDelete:
    def delete_collection(
        self,
        name: str
    ) -> None:
        """
        Delete a collection and all its data.

        This permanently removes the collection and all documents,
        embeddings, and metadata stored within it.

        Args:
            name: Name of the collection to delete

        Raises:
            ValueError: If collection doesn't exist

        Example:
            client.delete_collection("old-articles")
        """

        # delete from database
        self._db.delete_collection(name)
        # remove from cache
        if name in self._collections:
            del self._collections[name]


"""
Module containing the Utils class, which is used to do a bunch of stuff related to get operation in a collection.
"""

from typing import (
    Any,
    Dict,
    List
)

class Utils:
    @property
    def name(self) -> str:
        """
        Get collection name.
        """

        return self._name

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Get collection metadata.
        """

        return self._metadata
    
    def count(self) -> int:
        """
        Count the number of items in the collection.

        Returns:
            Number of items in the collection

        Example:
            print(f"Collection has {collection.count()} items")
        """

        return self._db.count(self._name)

    def peek(
        self,
        limit: int = 10
    ) -> Dict[str, List[Any]]:
        """
        Get a sample of items from the collection.

        Args:
            limit: Maximum number of items to return (default: 10)

        Returns:
            Dictionary with sample items

        Example:
            sample = collection.peek(5)
            print(f"Sample IDs: {sample['ids']}")
        """

        return self.get(limit=limit)


"""
Collection mixins API client module for Skypydb
"""

from skypydb.api.mixins.vector.collection.sysadd import SysAdd
from skypydb.api.mixins.vector.collection.sysget import SysGet
from skypydb.api.mixins.vector.collection.sysquery import SysQuery
from skypydb.api.mixins.vector.collection.sysupdate import SysUpdate
from skypydb.api.mixins.vector.collection.sysdelete import SysDelete
from skypydb.api.mixins.vector.collection.utils import Utils

__all__ = [
    "SysAdd",
    "SysGet",
    "SysQuery",
    "SysUpdate",
    "SysDelete",
    "Utils"
]


"""
Validation mixin module for Skypydb.
"""

from skypydb.security.mixins.validation.syscheck import SysCheck
from skypydb.security.mixins.validation.sysvalidation import SysValidation, validate_table_name, validate_column_name
from skypydb.security.mixins.validation.syssanitize import SysSanitize, sanitize_input

__all__ = [
    "SysCheck",
    "SysValidation",
    "validate_table_name",
    "validate_column_name",
    "SysSanitize",
    "sanitize_input"
]



"""
Module containing the SysGet class, which is used to get a table.
"""

from skypydb.errors import TableNotFoundError
from skypydb.table.table import Table
from skypydb.api.mixins.reactive.schema_loader import load_schema

class SysGet:
    def get_table(
        self,
        table_name: str
    ) -> Table:
        """
        Get an existing table by name.

        Args:
            table_name: Name of the table

        Returns:
            Table instance

        Raises:
            TableNotFoundError: If table doesn't exist

        Example:
            tables = client.create_table()
            users_table = tables["users"]
            
            # Later, get the table again:
            users = client.get_table("users")
        """

        if not self.db.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")
        return Table(self.db, table_name)

    def get_or_create_table(self) -> dict:
        """
        Get existing tables from schema or create them if missing.

        Returns:
            Dictionary mapping table names to Table instances

        Example:
            tables = client.get_or_create_table()
            users_table = tables["users"]
        """

        schema = load_schema()

        table_names = self.db.get_or_create_table(schema)
        return {
            name: Table(self.db, name)
            for name in table_names
        }


"""
Reactive mixins API client module for Skypydb
"""

from skypydb.api.mixins.reactive.syscreate import SysCreate
from skypydb.api.mixins.reactive.sysget import SysGet
from skypydb.api.mixins.reactive.sysdelete import SysDelete

__all__ = [
    "SysCreate",
    "SysGet",
    "SysDelete"
]


"""
Module containing the Utils class, which is used to make the embed class callable.
"""

from typing import List

class Utils:
    def __call__(
        self,
        texts: List[str]
    ) -> List[List[float]]:
        """
        Make the class callable for compatibility with other libraries.

        Args:
            texts: List of texts to embed

        Returns:
            List of embedding vectors
        """

        return self.embed(texts)



"""
Module containing the SysManager class, which is used to create an EncryptionManager instance.
"""

from typing import Optional
from skypydb.errors import EncryptionError
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from skypydb.security.mixins.encryption.syspassword import SysPassword

class SysManager:
    def __init__(
        self,
        encryption_key: Optional[str] = None,
        iterations: int = 100000,
        salt: Optional[bytes] = None
    ):
        if encryption_key is not None and not encryption_key.strip():
            raise EncryptionError("encryption_key must be a non-empty string")

        self.enabled = bool(encryption_key)
        self.iterations = iterations
        self._salt = salt
        self._key: Optional[bytes] = None
        self._password = SysPassword()
        if self.enabled:
            if encryption_key == "":
                raise EncryptionError("Encryption key must not be empty.")
            # derive a 256-bit key from the password
            assert encryption_key is not None  # type narrowing for type checker
            self._key = self._password._derive_key(encryption_key, salt=self._salt)
            self._aesgcm = AESGCM(self._key)

def create_encryption_manager(
    encryption_key: Optional[str] = None,
    salt: Optional[bytes] = None,
) -> SysManager:
    """
    Factory function to create an EncryptionManager instance.

    Args:
        encryption_key: Master encryption key. If None, encryption is disabled.
        salt: Required, non-empty salt for PBKDF2HMAC when encryption is enabled.

    Returns:
        EncryptionManager instance
    """

    return SysManager(encryption_key=encryption_key, salt=salt)



"""
Database link metadata utilities for dashboard auto-discovery.
"""

import struct
from pathlib import Path
from typing import (
    Dict,
    List,
    Optional,
    Tuple
)

class DatabaseLinker:
    """
    Manages database link metadata for automatic discovery and classification.

    The DatabaseLinker class handles the creation, reading, and discovery of binary
    metadata files that store information about database paths and their types.
    This enables dashboard applications to automatically discover and categorize
    databases without requiring explicit configuration.
    """

    def __init__(
        self,
        folder: str = "link",
        binary_file_type: Dict[str, str] = {"reactive": "reactivetype.bin", "vector": "vectortype.bin"},
        magic: bytes = b"SKYPYLINKER"
    ):
        """
        Initialize the DatabaseLinker with custom configuration.

        Args:
            folder: Name of the sidecar folder to store metadata files. This folder is created adjacent to database files. Defaults to "link".
            binary_file_type: Mapping from database type to metadata filename. Defaults to {"reactive": "reactivetype.bin", "vector": "vectortype.bin"}.
            magic: Magic bytes for file validation and identification. Must be 11 bytes. Defaults to b"SKYPYLINKER". Changing this affects discovery behavior.
        """

        self.folder = folder
        self.binary_file_type = binary_file_type
        self.magic = magic

    def discover_database_links(
        self,
        root: Optional[Path] = None
    ) -> List[Dict[str, str]]:
        """
        Recursively discover all database link metadata files from a root folder.

        This method searches for binary metadata files matching the pattern
        **/SKYPYLINKER/*.bin and decodes them to extract database information.

        Args:
            root: Root directory to start searching from. Defaults to the current working directory. If None, uses Path.cwd().

        Returns:
            List: List of dictionaries, each containing:
            type: Database type ("reactive" or "vector")
            path: Full path to the database file
            metadata_file: Full path to the metadata binary file
        """

        search_root = root or Path.cwd()
        discovered: List[Dict[str, str]] = []
        pattern = f"**/{self.magic}/*.bin"
        for metadata_file in search_root.glob(pattern):
            if not metadata_file.is_file():
                continue
            parsed_entries = self._read_link_metadata(metadata_file)
            discovered.extend(parsed_entries)
        return discovered

    def ensure_db_link_metadata(
        self,
        path: str,
        db_type: Optional[str] = None
    ) -> Dict[str, str]:
        """
        Create or update a sidecar link folder and binary metadata file for a database.

        This method creates the metadata directory structure and writes a binary
        metadata file containing the database path and type. If metadata already
        exists for the given database, the path is preserved and deduplicated.

        Args:
            path: Path to the database file. Can be absolute or relative.
            db_type: Explicit database type ("reactive" or "vector"). If not provided, the type will be auto-detected by inspecting the database schema.

        Returns:
            Dict[str, str]: Dictionary containing:
            type: The database type used
            path: The normalized absolute path to the database
        """

        resolved_db_path = self._resolve_db_path(path)
        resolved_type = db_type
        if resolved_type is None:
            raise ValueError(f"db_type must be specified ('reactive' or 'vector')")
        metadata_path = self._metadata_file_for_db(resolved_db_path, resolved_type)
        metadata_path.parent.mkdir(parents=True, exist_ok=True)

        existing_paths: List[str] = []
        if metadata_path.exists():
            try:
                raw = metadata_path.read_bytes()
                decoded = self._decode_binary_payload(raw)
                if decoded is not None and decoded[0] == resolved_type:
                    existing_paths = decoded[1]
            except OSError:
                existing_paths = []

        normalized = str(resolved_db_path)
        merged_paths = list(dict.fromkeys(existing_paths + [normalized]))

        payload = {
            "type": resolved_type,
            "path": normalized
        }
        metadata_path.write_bytes(
            self._encode_binary_payload(
                resolved_type,
                merged_paths
            )
        )
        return payload

    def _resolve_db_path(
        self,
        path: str
    ) -> Path:
        """
        Resolve a database path to an absolute Path object.

        Converts relative paths to absolute paths by joining with the current
        working directory, then resolves any symbolic links or path components.

        Args:
            path: Database path as a string, either absolute or relative.

        Returns:
            Path: Resolved absolute Path object pointing to the database file.
        """

        db_path = Path(path)
        return db_path if db_path.is_absolute() else (Path.cwd() / db_path).resolve()

    def _metadata_file_for_db(
        self,
        db_path: Path,
        db_type: str
    ) -> Path:
        """
        Get the path to the metadata file for a database.

        Constructs the metadata file path by creating a link folder in the
        same directory as the database and appending the appropriate binary
        filename based on the database type.

        Args:
            db_path: Path to the database file.
            db_type: Database type ("reactive" or "vector").

        Returns:
            Path: Full path to the metadata binary file.
        """

        link_dir = db_path.parent / self.folder
        return link_dir / self.binary_file_type[db_type]

    def _encode_binary_payload(
        self,
        db_type: str,
        db_paths: List[str]
    ) -> bytes:
        """
        Encode database metadata into a binary payload.

        Args:
            db_type: Database type ("reactive" or "vector").
            db_paths: List of database file paths to encode.

        Returns:
            bytes: Complete binary payload suitable for writing to a file.

        Raises:
            ValueError: If db_type is not "reactive" or "vector".
        """

        type_code = 1 if db_type == "reactive" else 2
        payload = bytearray()
        payload.extend(self.magic)
        payload.extend(struct.pack(">BI", type_code, len(db_paths)))
        for db_path in db_paths:
            path_bytes = db_path.encode("utf-8")
            payload.extend(struct.pack(">I", len(path_bytes)))
            payload.extend(path_bytes)
        return bytes(payload)

    def _decode_binary_payload(
        self,
        raw: bytes
    ) -> Optional[Tuple[str, List[str]]]:
        """
        Decode a binary payload into database metadata.

        Parses binary data created by _encode_binary_payload and extracts
        the database type and associated paths.

        Args:
            raw: Raw binary data from a metadata file.

        Returns:
            Optional[Tuple[str, List[str]]]: Tuple of (db_type, db_paths) if decoding is successful, None if the data is invalid or corrupted.
            db_type: "reactive" or "vector"
            db_paths: List of database file paths

        Validation:
            Checks magic header matches
            Verifies type code is valid (1 or 2)
            Ensures all path data fits within the payload
            Validates UTF-8 encoding for all paths
        """

        if len(raw) < len(self.magic) + 5 or not raw.startswith(self.magic):
            return None

        offset = len(self.magic)
        type_code, count = struct.unpack(">BI", raw[offset:offset + 5])
        db_type = "reactive" if type_code == 1 else "vector" if type_code == 2 else None
        if db_type is None:
            return None

        paths: List[str] = []
        cursor = offset + 5
        for _ in range(count):
            if cursor + 4 > len(raw):
                return None
            path_len = struct.unpack(">I", raw[cursor:cursor + 4])[0]
            cursor += 4
            if cursor + path_len > len(raw):
                return None
            try:
                db_path = raw[cursor:cursor + path_len].decode("utf-8")
            except UnicodeDecodeError:
                return None
            if db_path:
                paths.append(db_path)
            cursor += path_len
        if cursor != len(raw):
            return None
        return db_type, paths

    def _read_link_metadata(
        self,
        path: Path
    ) -> List[Dict[str, str]]:
        """
        Read and parse a database link metadata file.

        Reads a binary metadata file, decodes its contents, and returns a list
        of database entries with their metadata file path.

        Args:
            path: Path to the metadata binary file.

        Returns:
            List[Dict[str, str]]: List of database entries, each containing:
            type: Database type ("reactive" or "vector")
            path: Database file path
            metadata_file: Path to this metadata file
        """

        try:
            raw = path.read_bytes()
        except OSError:
            return []

        decoded = self._decode_binary_payload(raw)
        if decoded is None:
            return []

        db_type, db_paths = decoded
        entries: List[Dict[str, str]] = []
        for db_path in db_paths:
            entries.append(
                {
                    "type": db_type,
                    "path": db_path,
                    "metadata_file": str(path)
                }
            )
        return entries


"""
Module containing the SysValidate class, which is used to validate a row of data against a table definition.
"""

from typing import (
    Dict,
    Any
)
from skypydb.schema.values import Validator

class SysValidate:
    def __init__(
        self,
        columns: Dict[str, Validator]
    ):
        self.columns = columns

    def validate_row(
        self,
        row_data: Dict[str, Any]
    ) -> None:
        """
        Validate a row of data against this table definition.

        Args:
            row_data: Dictionary of column names to values

        Raises:
            ValueError: If validation fails or required columns are missing
        """

        # check all required columns are present
        for col_name, validator in self.columns.items():
            if col_name not in row_data:
                # check if the column is optional or required
                is_optional = getattr(validator, 'optional', False)
                if is_optional:
                    continue
                raise ValueError(
                    f"Missing required column: '{col_name}'"
                )
            value = row_data[col_name]
            if not validator.validate(value):
                raise ValueError(
                    f"Invalid value for column '{col_name}': "
                    f"expected {validator}, got {type(value).__name__}"
                )



"""
Module containing the SysGet class, which is used to get a collection.
"""

from typing import (
    Any,
    Dict,
    Optional,
    TYPE_CHECKING
)

if TYPE_CHECKING:
    from skypydb.api.collection import Collection

class SysGet:
    def get_collection(
        self,
        name: str
    ) -> "Collection":
        """
        Get an existing collection by name.

        Args:
            name: Name of the collection to retrieve

        Returns:
            Collection instance

        Raises:
            ValueError: If collection doesn't exist

        Example:
            collection = client.get_collection("articles")
        """

        # check if collection exists
        collection_info = self._db.get_collection(name)
        if collection_info is None:
            raise ValueError(f"Collection '{name}' not found")

        # return cached instance if available
        if name in self._collections:
            return self._collections[name]

        # create new collection instance
        from skypydb.api.collection import Collection
        collection = Collection(
            db=self._db,
            name=name,
            metadata=collection_info.get("metadata")
        )
        self._collections[name] = collection
        return collection

    def get_or_create_collection(
        self,
        name: str,
        metadata: Optional[Dict[str, Any]] = None
    ) -> "Collection":
        """
        Get an existing collection or create a new one.

        Args:
            name: Name of the collection
            metadata: Optional metadata (used only when creating)

        Returns:
            Collection instance

        Example:
            # Always works, whether collection exists or not
            collection = client.get_or_create_collection("articles")
        """

        # get or create in database
        collection_info = self._db.get_or_create_collection(name, metadata)

        # return cached instance if available
        if name in self._collections:
            return self._collections[name]

        # create new collection instance
        from skypydb.api.collection import Collection
        collection = Collection(
            db=self._db,
            name=name,
            metadata=collection_info.get("metadata")
        )
        self._collections[name] = collection
        return collection


"""
Skypydb - Open Source Reactive and Vector Embedding Database for Python.
"""

from skypydb.api.reactive_client import ReactiveClient
from skypydb.api.vector_client import VectorClient
from skypydb.api.collection import Collection
from skypydb.errors import (
    DatabaseError,
    InvalidSearchError,
    SkypydbError,
    TableAlreadyExistsError,
    TableNotFoundError
)
from skypydb.security import (
    EncryptionManager,
    EncryptionError,
    create_encryption_manager
)
from skypydb.embeddings import (
    OllamaEmbedding,
    OpenAIEmbedding,
    SentenceTransformerEmbedding,
    get_embedding_function
)

__version__ = "1.0.0"

__all__ = [
    "ReactiveClient",
    "VectorClient",
    "Collection",
    "SkypydbError",
    "DatabaseError",
    "TableNotFoundError",
    "TableAlreadyExistsError",
    "InvalidSearchError",
    "EncryptionManager",
    "EncryptionError",
    "create_encryption_manager",
    "OllamaEmbedding",
    "OpenAIEmbedding",
    "SentenceTransformerEmbedding",
    "get_embedding_function"
]



"""
Schema system for Skypydb.
"""

from typing import (
    List,
    Dict,
    Any,
    Optional
)
from .values import Validator
from skypydb.schema.mixins.schema.sysdef import SysDef
from skypydb.schema.mixins.schema.sysget import SysGet
from skypydb.schema.mixins.schema.sysschema import SysSchema
from skypydb.schema.mixins.schema.sysindex import SysIndex
from skypydb.schema.mixins.schema.sysvalidate import SysValidate

class TableDefinition(
    SysIndex,
    SysValidate,
    SysDef,
    SysGet,
    SysSchema
):
    """
    Definition of a table with columns and indexes.
    """

    def __init__(
        self,
        columns: Dict[str, Validator],
        table_name: Optional[str] = None
    ):
        """
        Initialize table definition.

        Args:
            columns: Dictionary mapping column names to validators
            table_name: Optional table name (can be used later)
        """

        self.columns = columns
        self.indexes: List[Dict[str, Any]] = []
        self.table_name = table_name



"""
Client API for Skypydb.
"""

import os
from typing import Optional
from skypydb.database.reactive_db import ReactiveDatabase
from skypydb.database.database_linker import DatabaseLinker
from skypydb.api.mixins.reactive import (
    SysCreate,
    SysGet,
    SysDelete
)

class ReactiveClient(
    SysCreate,
    SysGet,
    SysDelete
):
    """
    Reactive client for interacting with Skypydb.
    """

    def __init__(
        self,
        path: str = "./db/_generated/skypydb.db",
        encryption_key: Optional[str] = None,
        salt: Optional[bytes] = None,
        encrypted_fields: Optional[list] = None
    ):
        """
        Initialize Skypydb client.

        Args:
            path: Path to the database file. Defaults to ./db/_generated/skypydb.db
            encryption_key: Optional encryption key for data encryption at rest.
                           If provided, sensitive data will be encrypted.
                           Generate a secure key with: EncryptionManager.generate_key()
            salt: Optional salt for PBKDF2HMAC; must be non-empty and provided
                                 when encryption is enabled (i.e., when an encryption_key is set).
            encrypted_fields: Optional list of field names to encrypt.
                             If None and encryption is enabled, all fields except
                             'id' and 'created_at' will be encrypted.

        Example:
            # Without encryption
            client = skypydb.Client()

            # With encryption (all fields encrypted by default)
            from skypydb.security import EncryptionManager

            key = EncryptionManager.generate_key()

            client = skypydb.Client(encryption_key=key)

            # With encryption (specific fields only)
            client = skypydb.Client(
                encryption_key=key,
                encrypted_fields=["content", "email", "password"]
            )
        """

        # constant to define the path to the database file
        DB_PATH = path

        db_dir = os.path.dirname(DB_PATH)
        if db_dir and not os.path.exists(db_dir):
            os.makedirs(db_dir, exist_ok=True)

        self.database_linker = DatabaseLinker()

        self.path = DB_PATH
        self.db = ReactiveDatabase(
            DB_PATH,
            encryption_key=encryption_key,
            salt=salt,
            encrypted_fields=encrypted_fields
        )
        self.database_linker.ensure_db_link_metadata(DB_PATH, db_type="reactive")

    def close(self) -> None:
        """
        Close database connection.
        """

        self.db.close()



"""
Embedding function module.
"""

from skypydb.embeddings.mixins.embeddings_fn import EmbeddingsFn
from skypydb.embeddings.mixins.sysget import get_embedding_function
from skypydb.embeddings.mixins.utils import Utils

__all__ = [
    "EmbeddingsFn",
    "Utils",
    "get_embedding_function"
]



"""
Table API for managing data.
"""

from skypydb.database.reactive_db import ReactiveDatabase
from skypydb.errors import TableNotFoundError
from skypydb.table.mixins import (
    SysAdd,
    SysDelete,
    SysGet,
    SysSearch
)

class Table(
    SysAdd,
    SysDelete,
    SysGet,
    SysSearch
):
    """
    Represents a table in the database.
    """

    def __init__(
        self,
        db: ReactiveDatabase,
        table_name: str
    ):
        """
        Initialize table.

        Args:
            db: Database instance
            table_name: Name of the table
        """

        self.db = db
        self.table_name = table_name
        if not self.db.table_exists(table_name):
            raise TableNotFoundError(f"Table '{table_name}' not found")



"""
Schema loading utilities for the reactive client.
"""

import importlib.util
import os
from skypydb.schema import SysSchema


def load_schema() -> SysSchema:
    """
    Load schema from ./db/schema.py or package schema.py.
    """

    schema = None

    # first, try to load schema.py from the current working directory (./db/schema.py)
    cwd_schema_path = os.path.join(os.getcwd(), "db", "schema.py")
    if os.path.exists(cwd_schema_path):
        spec = importlib.util.spec_from_file_location("skypydb._user_schema", cwd_schema_path)
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            schema = getattr(module, "schema", None)
    # if not found in cwd, try to load from the installed package location
    if schema is None:
        package_root = os.path.dirname(os.path.dirname(__file__))
        schema_path = os.path.join(package_root, "schema.py")
        if os.path.exists(schema_path):
            spec = importlib.util.spec_from_file_location("skypydb._schema_file", schema_path)
            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                schema = getattr(module, "schema", None)
    if schema is None:
        raise ValueError(
            "No 'schema' object found in db/schema.py. "
            "Please create a schema.py file in ./db/schema.py with a schema definition "
            "using: schema = defineSchema({...})"
        )
    if not isinstance(schema, SysSchema):
        raise ValueError(
            f"Expected a Schema object, got {type(schema).__name__}"
        )
    return schema



"""
Module containing the SysUpdate class, which is used to update existing items in a collection.
"""

from typing import (
    Any,
    Dict,
    Optional,
    List
)

class SysUpdate:
    def update(
        self,
        ids: List[str],
        embeddings: Optional[List[List[float]]] = None,
        documents: Optional[List[str]] = None,
        metadatas: Optional[List[Dict[str, Any]]] = None
    ) -> None:
        """
        Update existing items in the collection.

        Args:
            ids: IDs of items to update
            embeddings: Optional new embeddings
            documents: Optional new documents (will be re-embedded)
            metadatas: Optional new metadata

        Example:
            # Update metadata only
            collection.update(
                ids=["doc1"],
                metadatas=[{"lang": "fr", "updated": True}]
            )
        """

        self._db.update(
            collection_name=self._name,
            ids=ids,
            embeddings=embeddings,
            documents=documents,
            metadatas=metadatas
        )


"""
Schema mixin module for Skypydb.
"""

from skypydb.schema.mixins.schema.sysindex import SysIndex
from skypydb.schema.mixins.schema.sysvalidate import SysValidate
from skypydb.schema.mixins.schema.sysdef import SysDef, defineTable, defineSchema
from skypydb.schema.mixins.schema.sysget import SysGet
from skypydb.schema.mixins.schema.sysschema import SysSchema

__all__ = [
    "SysIndex",
    "SysValidate",
    "SysDef",
    "SysGet",
    "SysSchema",
    "defineTable",
    "defineSchema"
]



"""
Module containing the SysCreate class, which is used to create a collection.
"""

from typing import (
    Any,
    Dict,
    Optional,
    TYPE_CHECKING
)

if TYPE_CHECKING:
    from skypydb.api.collection import Collection

class SysCreate:
    def create_collection(
        self,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
        get_or_create: bool = False
    ) -> "Collection":
        """
        Create a new collection.

        A collection is a named group of documents with their embeddings
        and metadata. Each document in the collection has a unique ID.

        Args:
            name: Unique name for the collection
            metadata: Optional metadata to attach to the collection
            get_or_create: If True, return existing collection if it exists

        Returns:
            Collection instance

        Raises:
            ValueError: If collection already exists and get_or_create is False

        Example:
            # Create a new collection
            collection = client.create_collection("articles")
            
            # Create or get existing
            collection = client.create_collection(
                "articles",
                get_or_create=True
            )
        """

        if get_or_create:
            # Ensure the collection exists, creating it if necessary.
            # We intentionally discard the returned instance here so that
            # this method can apply a consistent caching strategy via
            # self._collections below.
            self.get_or_create_collection(name, metadata)
        else:
            # Create collection in database
            self._db.create_collection(name, metadata)

        # Create and cache collection instance (or return cached one)
        _collection = self._collections.get(name)
        if _collection is None:
            from skypydb.api.collection import Collection
            _collection = Collection(
                db=self._db,
                name=name,
                metadata=metadata
            )
            self._collections[name] = _collection
        return _collection


"""
Module containing the Utils class, which is used to do a bunch of things with tables in the database.
"""

import sqlite3
import json
from datetime import datetime
from typing import (
    Optional,
    Dict,
    Any
)
from skypydb.security.validation import InputValidator
from skypydb.schema.schema import TableDefinition

class Utils:
    def __init__(
        self,
        path: Optional[str] = None,
        conn: Optional[sqlite3.Connection] = None
    ):
        if conn is not None:
            self.conn = conn
        elif path is not None:
            self.conn = sqlite3.connect(path, check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
        else:
            raise ValueError("Either path or conn must be provided")

    def get_table_config(
        self,
        table_name: str
    ) -> Optional[Dict[str, Any]]:
        """
        Retrieve a table's configuration from the system table.

        Args:
            table_name: Name of the table

        Returns:
            Configuration dictionary or None if not found
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        cursor = self.conn.cursor()

        cursor.execute(
            "SELECT config FROM _skypy_config WHERE table_name = ?", (table_name,)
        )
        row = cursor.fetchone()
        if row:
            return json.loads(row[0])
        return None

    def save_table_config(
        self,
        table_name: str,
        config: Dict[str, Any]
    ) -> None:
        """
        Save a table's configuration to the system table.

        Args:
            table_name: Name of the table
            config: Configuration dictionary for the table
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        # normalize config to ensure types are strings for JSON serialization
        normalized_config = self.normalize_config(config)

        cursor = self.conn.cursor()

        cursor.execute(
            """
            INSERT OR REPLACE INTO _skypy_config (table_name, config, created_at)
            VALUES (?, ?, ?)
            """,
            (table_name, json.dumps(normalized_config), datetime.now().isoformat()),
        )
        self.conn.commit()


    def normalize_config(
        self,
        config: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Normalize a table configuration so every column type is represented as a JSON-serializable value.

        For each entry in `config`:
        - If the value is a dict, returns a copy with its `"type"` field converted to one of the strings `"str"`, `"int"`, `"float"`, or `"bool"` when the type exactly matches the corresponding Python type; otherwise the type is converted to its string form. Other keys in the dict are preserved.
        - If the value is a list, it is preserved unchanged.
        - For any other value, the value is replaced by the normalized type string as described above.

        Parameters:
            config (Dict[str, Any]): Mapping of column names to type descriptors (type, dict, or list).

        Returns:
            Dict[str, Any]: Normalized configuration with JSON-serializable type representations.
        """

        def _normalize_type(t: Any) -> str:
            """
            Map a Python type object to a canonical string name used in stored table configurations.

            Parameters:
                t (Any): A type object or value; exact matches to the builtin types `str`, `int`, `float`, and `bool` produce their canonical names.

            Returns:
                A string: `'str'` if `t` is `str`, `'int'` if `t` is `int`, `'float'` if `t` is `float`, `'bool'` if `t` is `bool`, otherwise `str(t)`.
            """

            if t is str:
                return "str"
            if t is int:
                return "int"
            if t is float:
                return "float"
            if t is bool:
                return "bool"
            return str(t)

        normalized: Dict[str, Any] = {}
        for col_name, col_type in config.items():
            if isinstance(col_type, dict):
                normalized[col_name] = {
                    **col_type,
                    "type": _normalize_type(col_type.get("type", "str")),
                }
            elif isinstance(col_type, list):
                normalized[col_name] = col_type
            else:
                normalized[col_name] = _normalize_type(col_type)
        return normalized

    def table_def_to_config(
        self,
        table_def: TableDefinition
    ) -> Dict[str, Any]:
        """
        Convert a TableDefinition to a config dictionary for storage.

        Args:
            table_def: TableDefinition to convert

        Returns:
            Configuration dictionary
        """

        config = {}

        # convert validators to type strings
        for col_name, validator in table_def.columns.items():
            validator_repr = repr(validator)
            # map validator repr to config type
            # check for optional first before checking inner types
            if "value.optional(" in validator_repr:
                # extract the base type from optional
                if "value.string()" in validator_repr:
                    config[col_name] = {"type": "str", "optional": True}
                elif "value.int64()" in validator_repr:
                    config[col_name] = {"type": "int", "optional": True}
                elif "value.float64()" in validator_repr:
                    config[col_name] = {"type": "float", "optional": True}
                elif "value.boolean()" in validator_repr:
                    config[col_name] = {"type": "bool", "optional": True}
                else:
                    config[col_name] = {"type": "str", "optional": True}
            elif "value.string()" in validator_repr:
                config[col_name] = "str"
            elif "value.int64()" in validator_repr:
                config[col_name] = "int"
            elif "value.float64()" in validator_repr:
                config[col_name] = "float"
            elif "value.boolean()" in validator_repr:
                config[col_name] = "bool"
            else:
                config[col_name] = "str"  # default
        # add index information
        if table_def.indexes:
            config["_indexes"] = [
                {"name": idx["name"], "fields": idx["fields"]}
                for idx in table_def.indexes
            ]
        return config

    def delete_table_config(
        self,
        table_name: str
    ) -> None:
        """
        Delete a table's configuration from the system table.

        Args:
            table_name: Name of the table

        Raises:
            ValidationError: If input parameters are invalid
        """

        # validate table name
        table_name = InputValidator.validate_table_name(table_name)

        cursor = self.conn.cursor()

        cursor.execute("DELETE FROM _skypy_config WHERE table_name = ?", (table_name,))
        self.conn.commit()



"""
Values mixin module for Skypydb.
"""

from skypydb.schema.mixins.values.validator import Validator
from skypydb.schema.mixins.values.optionalvalidator import OptionalValidator
from skypydb.schema.mixins.values.booleanvalidator import BooleanValidator
from skypydb.schema.mixins.values.stringvalidator import StringValidator
from skypydb.schema.mixins.values.intvalidator import Int64Validator
from skypydb.schema.mixins.values.floatvalidator import Float64Validator

__all__ = [
    "Validator",
    "OptionalValidator",
    "BooleanValidator",
    "StringValidator",
    "Int64Validator",
    "Float64Validator"
]



"""
Module containing the Int64Validator class, which is used by the validator.
"""

from typing import Any
from skypydb.schema.mixins.values import Validator

class Int64Validator(Validator):
    """
    Validator for integer values.
    """

    def validate(
        self,
        value: Any
    ) -> bool:
        """
        Check if value is an integer.
        """

        return isinstance(value, int) and not isinstance(value, bool)

    def __repr__(self) -> str:
        return "value.int64()"
